# Generated by Django 2.2.12 on 2020-04-21 23:36

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("common", "0014_auto_20190226_1117"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="configuration",
            name="can_see_other_members",
        ),
        migrations.AlterField(
            model_name="configuration",
            name="language",
            field=models.CharField(
                blank=True,
                choices=[
                    ("af", "Afrikaans"),
                    ("ar", "Arabic"),
                    ("ast", "Asturian"),
                    ("az", "Azerbaijani"),
                    ("bg", "Bulgarian"),
                    ("be", "Belarusian"),
                    ("bn", "Bengali"),
                    ("br", "Breton"),
                    ("bs", "Bosnian"),
                    ("ca", "Catalan"),
                    ("cs", "Czech"),
                    ("cy", "Welsh"),
                    ("da", "Danish"),
                    ("de", "German"),
                    ("dsb", "Lower Sorbian"),
                    ("el", "Greek"),
                    ("en", "English"),
                    ("en-au", "Australian English"),
                    ("en-gb", "British English"),
                    ("eo", "Esperanto"),
                    ("es", "Spanish"),
                    ("es-ar", "Argentinian Spanish"),
                    ("es-co", "Colombian Spanish"),
                    ("es-mx", "Mexican Spanish"),
                    ("es-ni", "Nicaraguan Spanish"),
                    ("es-ve", "Venezuelan Spanish"),
                    ("et", "Estonian"),
                    ("eu", "Basque"),
                    ("fa", "Persian"),
                    ("fi", "Finnish"),
                    ("fr", "French"),
                    ("fy", "Frisian"),
                    ("ga", "Irish"),
                    ("gd", "Scottish Gaelic"),
                    ("gl", "Galician"),
                    ("he", "Hebrew"),
                    ("hi", "Hindi"),
                    ("hr", "Croatian"),
                    ("hsb", "Upper Sorbian"),
                    ("hu", "Hungarian"),
                    ("hy", "Armenian"),
                    ("ia", "Interlingua"),
                    ("id", "Indonesian"),
                    ("io", "Ido"),
                    ("is", "Icelandic"),
                    ("it", "Italian"),
                    ("ja", "Japanese"),
                    ("ka", "Georgian"),
                    ("kab", "Kabyle"),
                    ("kk", "Kazakh"),
                    ("km", "Khmer"),
                    ("kn", "Kannada"),
                    ("ko", "Korean"),
                    ("lb", "Luxembourgish"),
                    ("lt", "Lithuanian"),
                    ("lv", "Latvian"),
                    ("mk", "Macedonian"),
                    ("ml", "Malayalam"),
                    ("mn", "Mongolian"),
                    ("mr", "Marathi"),
                    ("my", "Burmese"),
                    ("nb", "Norwegian Bokmål"),
                    ("ne", "Nepali"),
                    ("nl", "Dutch"),
                    ("nn", "Norwegian Nynorsk"),
                    ("os", "Ossetic"),
                    ("pa", "Punjabi"),
                    ("pl", "Polish"),
                    ("pt", "Portuguese"),
                    ("pt-br", "Brazilian Portuguese"),
                    ("ro", "Romanian"),
                    ("ru", "Russian"),
                    ("sk", "Slovak"),
                    ("sl", "Slovenian"),
                    ("sq", "Albanian"),
                    ("sr", "Serbian"),
                    ("sr-latn", "Serbian Latin"),
                    ("sv", "Swedish"),
                    ("sw", "Swahili"),
                    ("ta", "Tamil"),
                    ("te", "Telugu"),
                    ("th", "Thai"),
                    ("tr", "Turkish"),
                    ("tt", "Tatar"),
                    ("udm", "Udmurt"),
                    ("uk", "Ukrainian"),
                    ("ur", "Urdu"),
                    ("vi", "Vietnamese"),
                    ("zh-hans", "Simplified Chinese"),
                    ("zh-hant", "Traditional Chinese"),
                ],
                max_length=5,
                null=True,
                verbose_name="Language",
            ),
        ),
    ]



# Generated by Django 2.2.1 on 2019-06-11 10:18

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("sepa", "0002_auto_20171013_1436")]

    operations = [
        migrations.AddField(
            model_name="membersepa",
            name="mandate_state",
            field=models.CharField(
                choices=[
                    ("inactive", "Inactive"),
                    ("active", "Active"),
                    ("bounced", "Bounced"),
                    ("rescinded", "Rescinded"),
                ],
                default="active",
                max_length=10,
                verbose_name="Mandate state",
            ),
        )
    ]



from collections import Counter

from django.contrib.auth import get_user_model
from django.db import models, transaction
from django.db.models import Prefetch
from django.urls import reverse
from django.utils.safestring import mark_safe

from byro.common.models import LogTargetMixin, log_call
from byro.documents.models import Document


class TransactionQuerySet(models.QuerySet):
    def with_balances(self):
        qs = self.annotate(
            balances_debit=models.Sum(
                models.Case(
                    models.When(
                        ~models.Q(bookings__debit_account=None), then="bookings__amount"
                    ),
                    default=0,
                    output_field=models.DecimalField(max_digits=8, decimal_places=2),
                )
            ),
            balances_credit=models.Sum(
                models.Case(
                    models.When(
                        ~models.Q(bookings__credit_account=None),
                        then="bookings__amount",
                    ),
                    default=0,
                    output_field=models.DecimalField(max_digits=8, decimal_places=2),
                )
            ),
        )
        return qs

    def unbalanced_transactions(self):
        return self.with_balances().exclude(balances_debit=models.F("balances_credit"))


class TransactionManager(models.Manager):
    def get_queryset(self):
        return TransactionQuerySet(self.model, using=self._db)

    def with_balances(self):
        return self.get_queryset().with_balances()

    def unbalanced_transactions(self):
        return self.get_queryset().unbalanced_transactions()

    @log_call(".created")
    def create(self, *args, **kwargs):
        return super().create(*args, **kwargs)


class Transaction(models.Model, LogTargetMixin):
    objects = TransactionManager()

    LOG_TARGET_BASE = "byro.bookkeeping.transaction"

    memo = models.CharField(max_length=1000, null=True)
    booking_datetime = models.DateTimeField(null=True)
    value_datetime = models.DateTimeField()
    modified = models.DateTimeField(auto_now=True)
    modified_by = models.ForeignKey(
        to=get_user_model(),
        on_delete=models.PROTECT,
        related_name="+",  # no related lookup
        null=True,
    )

    reverses = models.ForeignKey(
        to="Transaction",
        on_delete=models.PROTECT,
        related_name="reversed_by",
        null=True,
    )

    data = models.JSONField(null=True)

    documents = models.ManyToManyField(Document, through="DocumentTransactionLink")

    @log_call(".debit.created", log_on="self")
    def debit(self, account, *args, **kwargs):
        if self.credits.count() > 1 and self.debits.count() > 0:
            raise Exception(
                "Bug: Can't create debit. It should only be allowed to have either multiple debits or multiple credits."
            )

        return Booking.objects.create(
            transaction=self, debit_account=account, *args, **kwargs
        )

    @log_call(".credit.created", log_on="self")
    def credit(self, account, *args, **kwargs):
        if self.debits.count() > 1 and self.credits.count() > 0:
            raise Exception(
                "Bug: Can't create credit. It should only be allowed to have either multiple debits or multiple credits."
            )

        return Booking.objects.create(
            transaction=self, credit_account=account, *args, **kwargs
        )

    @transaction.atomic
    def reverse(self, value_datetime=None, *args, **kwargs):
        if "user_or_context" not in kwargs:
            raise TypeError(
                "You need to provide a 'user_or_context' named parameter which indicates the responsible user (a User model object), request (a View instance or HttpRequest object), or generic context (a str)."
            )
        user_or_context = kwargs.pop("user_or_context")
        user = kwargs.pop("user", None)

        t = Transaction.objects.create(
            value_datetime=value_datetime or self.value_datetime,
            reverses=self,
            user_or_context=user_or_context,
            user=user,
            *args,
            **kwargs,
        )
        for b in self.bookings.all():
            if b.credit_account:
                t.debit(
                    account=b.credit_account,
                    amount=b.amount,
                    member=b.member,
                    user_or_context=user_or_context,
                    user=user,
                )
            elif b.debit_account:
                t.credit(
                    account=b.debit_account,
                    amount=b.amount,
                    member=b.member,
                    user_or_context=user_or_context,
                    user=user,
                )
        t.save()
        self.log(user_or_context, ".reversed", user=user, reversed_by=t)

        return t

    @property
    def debits(self):
        return self.bookings.exclude(debit_account=None)

    @property
    def credits(self):
        return self.bookings.exclude(credit_account=None)

    @property
    def balances(self):
        balances = {
            "debit": self.debits.aggregate(total=models.Sum("amount"))["total"] or 0,
            "credit": self.credits.aggregate(total=models.Sum("amount"))["total"] or 0,
        }
        return balances

    @property
    def is_read_only(self):
        """Advisory property: don't modify this Transaction or its Bookings."""
        # Future proof: For now, don't modify balanced transactions
        return self.is_balanced

    @property
    def is_balanced(self):
        if hasattr(self, "balances_debit"):
            return self.balances_debit == self.balances_credit
        else:
            balances = self.balances
            return balances["debit"] == balances["credit"]

    def find_memo(self):
        if self.memo:
            return self.memo

        if hasattr(self, "cached_bookings"):
            for booking in self.cached_bookings:
                if booking.memo:
                    return booking.memo

        booking = self.bookings.exclude(memo=None).first()
        if booking:
            return booking.memo

        return None

    @transaction.atomic
    def process_transaction(self):
        """Collects responses to the signal `process_transaction`. Re-raises
        received Exceptions.

        Returns the number of receivers that augmented the Transaction.
        """
        from byro.bookkeeping.signals import process_transaction

        response_counter = Counter()
        this_counter = Counter("dummy")

        while (
            not response_counter or response_counter.most_common(1)[0][1] < 5
        ) and sum(this_counter.values()) > 0:
            responses = process_transaction.send_robust(sender=self)
            this_counter = Counter(
                receiver for receiver, response in responses if response
            )

            for _receiver, response in responses:
                if isinstance(response, Exception):
                    raise response

            response_counter += this_counter

        if sum(response_counter.values()) < 1:
            raise Exception("No plugin tried to augment the transaction.")

        response_counter += Counter()  # Remove zero and negative elements
        return len(response_counter)

    def __str__(self):
        return "Transaction(pk={}, memo={!r}, value_datetime={!r}{})".format(
            self.pk,
            self.find_memo(),
            self.value_datetime.isoformat(),
            f", reverses={self.reverses}" if self.reverses else "",
        )

    def get_absolute_url(self):
        return reverse("office:finance.transactions.detail", kwargs={"pk": self.pk})

    def get_object_icon(self):
        return mark_safe('<i class="fa fa-money"></i> ')


class DocumentTransactionLink(models.Model):
    document = models.ForeignKey(Document, on_delete=models.CASCADE)
    transaction = models.ForeignKey(Transaction, on_delete=models.CASCADE)


class BookingsQuerySet(models.QuerySet):
    def with_transaction_balances(self):
        qs = self.annotate(
            transaction_balances_debit=models.Sum(
                models.Case(
                    models.When(
                        ~models.Q(transaction__bookings__debit_account=None),
                        then="transaction__bookings__amount",
                    ),
                    default=0,
                    output_field=models.DecimalField(max_digits=8, decimal_places=2),
                )
            ),
            transaction_balances_credit=models.Sum(
                models.Case(
                    models.When(
                        ~models.Q(transaction__bookings__credit_account=None),
                        then="transaction__bookings__amount",
                    ),
                    default=0,
                    output_field=models.DecimalField(max_digits=8, decimal_places=2),
                )
            ),
        )
        return qs

    def with_transaction_data(self):
        qs = self.with_transaction_balances()
        qs = qs.select_related(
            "member", "transaction", "credit_account", "debit_account"
        )
        qs = qs.prefetch_related(
            Prefetch("transaction__bookings", to_attr="cached_bookings"),
            "transaction__cached_bookings__credit_account",
            "transaction__cached_bookings__debit_account",
            "transaction__cached_bookings__member",
        )
        return qs


class Booking(models.Model):
    objects = BookingsQuerySet.as_manager()

    memo = models.CharField(max_length=1000, null=True)

    booking_datetime = models.DateTimeField(null=True)
    modified = models.DateTimeField(auto_now=True)
    modified_by = models.ForeignKey(
        to=get_user_model(),
        on_delete=models.PROTECT,
        related_name="+",  # no related lookup
        null=True,
    )

    transaction = models.ForeignKey(
        to="Transaction", related_name="bookings", on_delete=models.PROTECT
    )
    amount = models.DecimalField(max_digits=8, decimal_places=2)
    debit_account = models.ForeignKey(
        to="bookkeeping.Account",
        related_name="debits",
        on_delete=models.PROTECT,
        null=True,
    )
    credit_account = models.ForeignKey(
        to="bookkeeping.Account",
        related_name="credits",
        on_delete=models.PROTECT,
        null=True,
    )
    member = models.ForeignKey(
        to="members.Member",
        related_name="bookings",
        on_delete=models.PROTECT,
        null=True,
    )

    importer = models.CharField(null=True, max_length=500)
    data = models.JSONField(null=True)
    source = models.ForeignKey(
        to="bookkeeping.RealTransactionSource",
        on_delete=models.SET_NULL,
        related_name="bookings",
        null=True,
    )

    def __str__(self):
        return "{booking_type} {account} {amount} {memo}".format(
            booking_type="debit" if self.debit_account else "credit",
            account=self.debit_account or self.credit_account,
            amount=self.amount,
            memo=self.memo,
        )

    def save(self, *args, **kwargs):
        if self.debit_account_id and self.credit_account_id:
            raise Exception("Must be either credit or debit transaction, not both!")
        if not self.debit_account_id and not self.credit_account_id:
            raise Exception("Must be either credit or debit transaction, not neither!")
        return super().save(*args, **kwargs)

    def find_memo(self):
        if self.memo:
            return self.memo
        return self.transaction.find_memo()

    @property
    def counter_bookings(self):
        if hasattr(self.transaction, "cached_bookings"):
            # Was prefetched with with_transaction_data()
            if self.debit_account:
                return [b for b in self.transaction.cached_bookings if b.credit_account]
            elif self.credit_account:
                return [b for b in self.transaction.cached_bookings if b.debit_account]
        else:
            if self.debit_account:
                return self.transaction.credits
            elif self.credit_account:
                return self.transaction.debits
        return None



from django.db import models, transaction

from byro.common.models import LogTargetMixin
from byro.common.models.auditable import Auditable
from byro.common.models.choices import Choices

from .transaction import Transaction


class SourceState(Choices):
    NEW = "new"
    PROCESSING = "processing"
    PROCESSED = "processed"
    FAILED = "failed"


class RealTransactionSource(Auditable, models.Model, LogTargetMixin):
    source_file = models.FileField(upload_to="transaction_uploads/")
    state = models.CharField(
        default=SourceState.NEW,
        choices=SourceState.choices,
        max_length=SourceState.max_length,
    )

    @transaction.atomic
    def process(self):
        """Collects responses to the signal `process_csv_upload`. Raises an
        exception if multiple results were found, and re-raises received
        Exceptions.

        Returns a list of one or more Transaction objects if no
        Exception was raised.
        """
        self.state = SourceState.PROCESSING
        self.save()
        from byro.bookkeeping.signals import process_csv_upload

        responses = process_csv_upload.send_robust(sender=self)
        if len(responses) > 1:
            self.state = SourceState.FAILED
            self.save()
            raise Exception(
                "More than one plugin tried to process the CSV upload: {}".format(
                    [r[0].__module__ + "." + r[0].__name__ for r in responses]
                )
            )
        if len(responses) < 1:
            self.state = SourceState.FAILED
            self.save()
            raise Exception("No plugin tried to process the CSV upload.")
        receiver, response = responses[0]

        if isinstance(response, Exception):
            self.state = SourceState.FAILED
            self.save()
            raise response

        for t in Transaction.objects.filter(bookings__source=self):
            try:
                t.process_transaction()
            except Exception as e:
                print(e)  # FIXME

        self.state = SourceState.PROCESSED
        self.save()
        return response

    @property
    def transactions(self):
        """Get all transactions."""
        return Transaction.objects.filter(bookings__source=self)



# Generated by Django 1.11.13 on 2018-06-17 19:26

from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("members", "0009_auto_20180512_1810"),
        ("bookkeeping", "0011_auto_20180303_1745"),
    ]

    operations = [
        migrations.CreateModel(
            name="AccountTag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=300, unique=True)),
                ("description", models.CharField(max_length=1000, null=True)),
            ],
        ),
        migrations.CreateModel(
            name="Booking",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("memo", models.CharField(max_length=1000, null=True)),
                ("booking_datetime", models.DateTimeField(null=True)),
                ("modified", models.DateTimeField(auto_now=True)),
                ("amount", models.DecimalField(decimal_places=2, max_digits=8)),
                ("importer", models.CharField(max_length=500, null=True)),
                ("data", django.contrib.postgres.fields.jsonb.JSONField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name="Transaction",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("memo", models.CharField(max_length=1000, null=True)),
                ("booking_datetime", models.DateTimeField(null=True)),
                ("value_datetime", models.DateTimeField()),
                ("modified", models.DateTimeField(auto_now=True)),
                ("data", django.contrib.postgres.fields.jsonb.JSONField(null=True)),
                (
                    "modified_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "reverses",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="reversed_by",
                        to="bookkeeping.Transaction",
                    ),
                ),
            ],
        ),
        migrations.AlterField(
            model_name="account",
            name="account_category",
            field=models.CharField(
                choices=[
                    ("member_donation", "Donation account"),
                    ("member_fees", "Membership fee account"),
                    ("asset", "Asset account"),
                    ("liability", "Liability account"),
                    ("income", "Income account"),
                    ("expense", "Expense account"),
                    ("equity", "Equity account"),
                ],
                max_length=15,
            ),
        ),
        migrations.AddField(
            model_name="booking",
            name="credit_account",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="credits",
                to="bookkeeping.Account",
            ),
        ),
        migrations.AddField(
            model_name="booking",
            name="debit_account",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="debits",
                to="bookkeeping.Account",
            ),
        ),
        migrations.AddField(
            model_name="booking",
            name="member",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="bookings",
                to="members.Member",
            ),
        ),
        migrations.AddField(
            model_name="booking",
            name="modified_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="booking",
            name="source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="bookings",
                to="bookkeeping.RealTransactionSource",
            ),
        ),
        migrations.AddField(
            model_name="booking",
            name="transaction",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="bookings",
                to="bookkeeping.Transaction",
            ),
        ),
        migrations.AddField(
            model_name="account",
            name="tags",
            field=models.ManyToManyField(
                related_name="accounts", to="bookkeeping.AccountTag"
            ),
        ),
    ]



from django.apps import AppConfig
from django.conf import settings
from django.db.models.signals import post_save


class CommonConfig(AppConfig):
    name = "byro.common"


def user_save_receiver(sender, instance, created, **kwargs):
    if created:
        from byro.common.models import LogEntry

        LogEntry.objects.create(
            content_object=instance,
            action_type="byro.common.user.created",
            data={
                "source": "User creation checkpoint",
                "flags": {
                    "active": instance.is_active,
                    "superuser": instance.is_superuser,
                    "staff": instance.is_staff,
                },
            },
        )


post_save.connect(user_save_receiver, sender=settings.AUTH_USER_MODEL)



# Generated by Django 3.1 on 2020-08-20 20:18

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("members", "0012_split_names"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="member",
            options={"ordering": (("direct_address_name", "name"),)},
        ),
    ]



from django import template

register = template.Library()


@register.simple_tag
def url_replace(request, field, value):
    queryparams = request.GET.copy()
    queryparams[field] = str(value)
    return queryparams.urlencode()



# Generated by Django 2.0 on 2017-12-06 19:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("profile", "0003_auto_20171012_2032")]

    operations = [
        migrations.AlterField(
            model_name="memberprofile",
            name="birth_date",
            field=models.DateField(null=True, verbose_name="Birth date"),
        ),
        migrations.AlterField(
            model_name="memberprofile",
            name="nick",
            field=models.CharField(max_length=200, null=True, verbose_name="Nick"),
        ),
        migrations.AlterField(
            model_name="memberprofile",
            name="phone_number",
            field=models.CharField(
                blank=True, max_length=32, null=True, verbose_name="Phone number"
            ),
        ),
    ]



# Generated by Django 1.11.4 on 2017-08-12 10:46

import byro.common.models.auditable
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Member",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("email", models.EmailField(max_length=254)),
            ],
            bases=(byro.common.models.auditable.Auditable, models.Model),
        )
    ]



from django import forms
from django.conf import settings
from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.shortcuts import redirect, reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import CreateView, ListView, UpdateView, View
from i18nfield.forms import I18nModelForm

from byro.mails.models import EMail, MailTemplate
from byro.members.models import Member


class RestrictedLanguagesI18nModelForm(I18nModelForm):
    def __init__(self, *args, **kwargs):
        if "locales" not in kwargs:
            from byro.common.models import Configuration

            config = Configuration.get_solo()
            kwargs["locales"] = [config.language or settings.LANGUAGE_CODE]
        return super().__init__(*args, **kwargs)


class MailSpecialToFormClass(forms.ModelForm):
    class Meta:
        model = EMail
        form = RestrictedLanguagesI18nModelForm
        fields = ["to", "reply_to", "cc", "bcc", "subject", "text"]

    to_type = forms.ChoiceField(
        choices=[
            ("addr", _("Specific address")),
            ("member", _("Member")),
            ("all", _("All members")),
        ],
        widget=forms.RadioSelect,
        initial="addr",
    )

    to_member = forms.ModelChoiceField(
        Member.objects.filter(email__isnull=False).exclude(email=""),
        required=False,
        empty_label=None,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._decode_special(self.initial)
        self.fields[
            "to"
        ].required = (
            False  # FIXME Needs to be re-added in case no special mode is active
        )
        self.order_fields(["to_type", "to_member"])

    @staticmethod
    def _decode_special(d):
        if not d:
            return
        if "to" in d:
            to = d["to"]
            d["to_type"] = "addr"
            if to.startswith("special:"):
                if to.startswith("special:all"):
                    d["to_type"] = "all"
                    del d["to"]
                elif to.startswith("special:member:"):
                    d["to_type"] = "member"
                    d["to_member"] = Member.objects.get(pk=to.split(":", 2)[2])
                    del d["to"]

    def _encode_special(self):
        if self.cleaned_data["to_type"] == "all":
            self.cleaned_data["to"] = "special:all"
        elif self.cleaned_data["to_type"] == "member":
            self.cleaned_data["to"] = "special:member:{}".format(
                self.cleaned_data["to_member"].pk
            )
        self.initial["to_type"] = None
        self.initial["to_member"] = None
        del self.cleaned_data["to_type"]
        del self.cleaned_data["to_member"]
        self.instance.to = self.cleaned_data["to"]
        if not self.cleaned_data["to"]:
            self.add_error("to", _("Field cannot be empty"))
            self.initial["to_type"] = "addr"

    def clean(self):
        self._encode_special()
        return super().clean()


class MailSendMixin:
    def form_valid(self, form):
        if form.instance.sent:
            raise forms.ValidationError(
                _(
                    "This mail has been sent already, and cannot be modified. Copy it to a draft instead!"
                )
            )
        result = super().form_valid(form)
        if form.data.get("action", "save") == "send":
            form.instance.send()
            messages.success(
                self.request, _("Your changes have been saved and the email was sent.")
            )
        else:
            messages.success(self.request, _("Your changes have been saved."))
        return result


class MailDetail(MailSendMixin, UpdateView):
    queryset = EMail.objects.all()
    template_name = "office/mails/detail.html"
    context_object_name = "mail"
    form_class = MailSpecialToFormClass
    success_url = "/mails/outbox"

    def get_form(self, *args, **kwargs):
        form = super().get_form(*args, **kwargs)
        if form.instance.sent:
            for field in form.fields.values():
                field.disabled = True
        return form


class MailCopy(View):
    def get_object(self):
        return EMail.objects.get(pk=self.kwargs["pk"])

    def dispatch(self, request, *args, **kwargs):
        new_mail = self.get_object().copy_to_draft()
        messages.success(request, _("Here is your new mail draft!"))
        return redirect(reverse("office:mails.mail.view", kwargs={"pk": new_mail.pk}))


class OutboxQueryset:
    def get_queryset(self):
        qs = EMail.objects.filter(sent__isnull=True)
        if "pk" in self.kwargs:
            return qs.filter(pk=self.kwargs["pk"])
        return qs


class OutboxList(OutboxQueryset, ListView):
    template_name = "office/mails/outbox.html"
    context_object_name = "mails"


class OutboxPurge(OutboxQueryset, View):
    def dispatch(self, request, *args, **kwargs):
        qs = self.get_queryset()
        length = len(qs)
        qs.delete()
        if length > 1:
            message = _("{count} mails have been deleted.").format(count=length)
        elif length == 1:
            message = "The mail has been deleted."
        else:
            message = "No mail has been deleted."
        messages.success(request, message)
        return redirect(reverse("office:mails.outbox.list"))


class OutboxSend(OutboxQueryset, View):
    def dispatch(self, request, *args, **kwargs):
        qs = self.get_queryset()
        length = len(qs)
        for mail in qs:
            mail.send()
        if length > 1:
            message = _("{count} mails have been sent.").format(count=length)
        elif length == 1:
            message = _("The mail has been sent.")
        else:
            message = _("No mail has been sent.")
        messages.success(request, message)
        return redirect(reverse("office:mails.outbox.list"))


class SentMail(ListView):
    queryset = EMail.objects.filter(sent__isnull=False).order_by("-sent")
    template_name = "office/mails/sent.html"
    context_object_name = "mails"


class TemplateList(ListView):
    queryset = MailTemplate.objects.all()
    template_name = "office/mails/templates.html"
    context_object_name = "templates"


MAIL_TEMPLATE_FORM_CLASS = forms.modelform_factory(
    MailTemplate,
    form=RestrictedLanguagesI18nModelForm,
    fields=["subject", "text", "reply_to", "bcc"],
)


class TemplateDetail(UpdateView):
    queryset = MailTemplate.objects.all()
    template_name = "office/mails/template_detail.html"
    context_object_name = "template"
    success_url = "/mails/templates"
    form_class = MAIL_TEMPLATE_FORM_CLASS


class TemplateCreate(CreateView):
    model = MailTemplate
    template_name = "office/mails/template_detail.html"
    context_object_name = "template"
    success_url = "/mails/templates"
    form_class = MAIL_TEMPLATE_FORM_CLASS


class Compose(SuccessMessageMixin, MailSendMixin, CreateView):
    model = MailTemplate
    template_name = "office/mails/compose.html"
    context_object_name = "template"
    success_url = "/mails/outbox"
    form_class = MailSpecialToFormClass

    def get_initial(self):
        return {k: v for (k, v) in self.request.GET.items()}


class TemplateDelete(View):  # TODO
    pass



import os
from contextlib import suppress
from pathlib import Path
from urllib.parse import urlparse

from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _
from pkg_resources import iter_entry_points

from byro.common.settings.config import build_config
from byro.common.settings.utils import log_initial

config, config_files = build_config()
CONFIG = config

##
# This settings file is rather lengthy. It follows this structure:
# Directories, Apps, Url, Security, Databases, Logging, Email, Caching (and Sessions)
# I18n, Auth, Middleware, Templates and Staticfiles, External Apps
#
# Search for "## {AREA} SETTINGS" to navigate this file
##

DEBUG = config.getboolean("site", "debug")

## DIRECTORY SETTINGS
BASE_DIR = config.get("filesystem", "base")
DATA_DIR = config.get(
    "filesystem",
    "data",
    fallback=os.environ.get("BYRO_DATA_DIR", os.path.join(BASE_DIR, "data")),
)
LOG_DIR = config.get("filesystem", "logs", fallback=os.path.join(DATA_DIR, "logs"))
MEDIA_ROOT = config.get("filesystem", "media", fallback=os.path.join(DATA_DIR, "media"))
STATIC_ROOT = config.get(
    "filesystem", "static", fallback=os.path.join(BASE_DIR, "static.dist")
)

for directory in (BASE_DIR, LOG_DIR, STATIC_ROOT, MEDIA_ROOT):
    os.makedirs(directory, exist_ok=True)


## APP SETTINGS
INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "compressor",
    "bootstrap4",
    "djangoformsetjs",
    "solo.apps.SoloAppConfig",
    "django_select2",
    "byro.common.apps.CommonConfig",
    "byro.bookkeeping.apps.BookkeepingConfig",
    "byro.documents.apps.DocumentsConfig",
    "byro.mails.apps.MailsConfig",
    "byro.members.apps.MemberConfig",
    "byro.office.apps.OfficeConfig",
    "byro.public.apps.PublicConfig",
    "byro.plugins.profile.ProfilePluginConfig",
    "byro.plugins.sepa.SepaPluginConfig",
    "annoying",
]

PLUGINS = []
for entry_point in iter_entry_points(group="byro.plugin", name=None):
    PLUGINS.append(entry_point.module_name)
    INSTALLED_APPS.append(entry_point.module_name)

## URL SETTINGS
SITE_URL = config.get("site", "url", fallback="http://localhost")
INTERNAL_IPS = ["127.0.0.1", "::1", "localhost"]
ALLOWED_HOSTS = [urlparse(SITE_URL).hostname]
if DEBUG:
    ALLOWED_HOSTS += INTERNAL_IPS
ROOT_URLCONF = "byro.urls"
STATIC_URL = "/static/"
MEDIA_URL = "/media/"


## SECURITY SETTINGS
X_FRAME_OPTIONS = "DENY"
SECURE_BROWSER_XSS_FILTER = True
SECURE_CONTENT_TYPE_NOSNIFF = True
CSRF_COOKIE_NAME = "byro_csrftoken"
CSRF_TRUSTED_ORIGINS = [urlparse(SITE_URL).hostname]
SESSION_COOKIE_NAME = "byro_session"
SESSION_COOKIE_SECURE = config.getboolean(
    "site", "https", fallback=SITE_URL.startswith("https:")
)
if config.has_option("site", "secret"):
    SECRET_KEY = config.get("site", "secret")
else:
    SECRET_FILE = os.path.join(DATA_DIR, ".secret")
    if os.path.exists(SECRET_FILE):
        with open(SECRET_FILE) as f:
            SECRET_KEY = f.read().strip()
    else:
        chars = "abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)"
        SECRET_KEY = get_random_string(50, chars)
        with open(SECRET_FILE, "w") as f:
            os.chmod(SECRET_FILE, 0o600)
            if hasattr(os, "chown"):
                os.chown(SECRET_FILE, os.getuid(), os.getgid())
            f.write(SECRET_KEY)


## DATABASE SETTINGS
db_name = config.get("database", "name", fallback=os.path.join(DATA_DIR, "db.sqlite3"))
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends."
        + config.get("database", "engine", fallback="postgresql"),
        "NAME": db_name,
        "USER": config.get("database", "user"),
        "PASSWORD": config.get("database", "password"),
        "HOST": config.get("database", "host"),
        "PORT": config.get("database", "port"),
        "CONN_MAX_AGE": 120,
    }
}
if os.getenv("TRAVIS"):
    DATABASES["default"]["USER"] = "postgres"
    DATABASES["default"]["HOST"] = "localhost"
    DATABASES["default"]["PASSWORD"] = ""

# for docker-compose development
if os.getenv("DEVELOPMENT"):
    DATABASES["default"]["USER"] = "byro"
    DATABASES["default"]["HOST"] = "db"
    DATABASES["default"]["PASSWORD"] = "byro"

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

## LOGGING SETTINGS
loglevel = "DEBUG" if DEBUG else "INFO"
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "format": "%(levelname)s %(asctime)s %(name)s %(module)s %(message)s"
        }
    },
    "handlers": {
        "console": {
            "level": loglevel,
            "class": "logging.StreamHandler",
            "formatter": "default",
        },
        "file": {
            "level": loglevel,
            "class": "logging.FileHandler",
            "filename": os.path.join(LOG_DIR, "byro.log"),
            "formatter": "default",
        },
    },
    "loggers": {
        "": {"handlers": ["file", "console"], "level": loglevel, "propagate": True},
        "django.request": {
            "handlers": ["file", "console"],
            "level": loglevel,
            "propagate": True,
        },
        "django.security": {
            "handlers": ["file", "console"],
            "level": loglevel,
            "propagate": True,
        },
        "django.db.backends": {
            "handlers": ["file", "console"],
            "level": "INFO",  # Do not output all the queries
            "propagate": True,
        },
    },
}

email_level = config.get("logging", "email_level", fallback="ERROR") or "ERROR"
emails = config.get("logging", "email", fallback="").split(",")
MANAGERS = ADMINS = [(email, email) for email in emails if email]
if ADMINS:
    LOGGING["handlers"]["mail_admins"] = {
        "level": email_level,
        "class": "django.utils.log.AdminEmailHandler",
    }


## EMAIL SETTINGS
MAIL_FROM = SERVER_EMAIL = DEFAULT_FROM_EMAIL = config.get("mail", "from")
if DEBUG:
    EMAIL_BACKEND = "django.core.mail.backends.console.EmailBackend"
else:
    EMAIL_HOST = config.get("mail", "host")
    EMAIL_PORT = config.get("mail", "port")
    EMAIL_HOST_USER = config.get("mail", "user")
    EMAIL_HOST_PASSWORD = config.get("mail", "password")
    EMAIL_USE_TLS = config.getboolean("mail", "tls")
    EMAIL_USE_SSL = config.getboolean("mail", "ssl")


## I18N SETTINGS
USE_I18N = True
USE_L10N = True
USE_TZ = True
LANGUAGES = [
    ("en", _("English")),
    ("de", _("German")),
    ("pt-br", _("Brazilian Portuguese")),
]
LANGUAGES_NATURAL_NAMES = [
    ("en", "English"),
    ("de", "Deutsch"),
    ("pt-br", "Português Brasileiro"),
]
LOCALE_PATHS = (Path(__file__).resolve().parent / "locale",)
FORMAT_MODULE_PATH = ["byro.common.formats"]
TIME_ZONE = config.get("locale", "time_zone")
LANGUAGE_CODE = "en"
DEFAULT_LANGUAGE = config.get("locale", "language_code") or "de"


## AUTHENTICATION SETTINGS
AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]


## MIDDLEWARE SETTINGS
MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "whitenoise.middleware.WhiteNoiseMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "byro.common.middleware.PermissionMiddleware",
    "byro.common.middleware.SettingsMiddleware",
]


## TEMPLATE AND STATICFILES SETTINGS
template_loaders = (
    "django.template.loaders.filesystem.Loader",
    "django.template.loaders.app_directories.Loader",
)
if not DEBUG:
    template_loaders = (("django.template.loaders.cached.Loader", template_loaders),)
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.template.context_processors.static",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "byro.common.context_processors.byro_information",
                "byro.common.context_processors.sidebar_information",
            ],
            "loaders": template_loaders,
        },
    }
]

STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    "compressor.finders.CompressorFinder",
)
STATICFILES_DIRS = [os.path.join(BASE_DIR, "byro", "static")]
STATICFILES_STORAGE = "whitenoise.storage.CompressedManifestStaticFilesStorage"


## EXTERNAL APP SETTINGS
with suppress(ImportError):
    import django_extensions  # noqa

    INSTALLED_APPS.append("django_extensions")

with suppress(ImportError):
    import django_securebox

    INSTALLED_APPS.append("django_securebox")
    MIDDLEWARE.insert(2, "django_securebox.middleware.SecureBoxMiddleware")

with suppress(ImportError):
    import debug_toolbar

    INSTALLED_APPS.append("debug_toolbar")
    MIDDLEWARE.insert(2, "debug_toolbar.middleware.DebugToolbarMiddleware")


COMPRESS_ENABLED = COMPRESS_OFFLINE = not DEBUG
COMPRESS_PRECOMPILERS = (("text/x-scss", "django_libsass.SassCompiler"),)
COMPRESS_FILTERS = {
    "js": ["compressor.filters.jsmin.rJSMinFilter"],
    "css": (
        # CssAbsoluteFilter is incredibly slow, especially when dealing with our _flags.scss
        # However, we don't need it if we consequently use the static() function in Sass
        # 'compressor.filters.css_default.CssAbsoluteFilter',
        "compressor.filters.cssmin.CSSCompressorFilter",
    ),
}
SELECT2_JS = ""
SELECT2_CSS = ""
SELECT2_I18N_PATH = "/static/vendored/select2/js/i18n"

with suppress(ImportError):
    from .local_settings import *

    print(
        "You are using the deprecated local_settings.py – Please move to the byro.cfg format."
    )

WSGI_APPLICATION = "byro.wsgi.application"
log_initial(
    debug=DEBUG,
    config_files=config_files,
    db_name=db_name,
    LOG_DIR=LOG_DIR,
    plugins=PLUGINS,
)



from contextlib import suppress
from enum import Enum

from annoying.fields import AutoOneToOneField
from django.db import models
from django.utils.translation import gettext_lazy as _
from localflavor.generic.models import BICField, IBANField
from schwifty import BIC, IBAN

from byro.common.models.auditable import Auditable


class SepaDirectDebitState(Enum):
    OK = _("OK")
    NO_IBAN = _("No IBAN")
    INVALID_IBAN = _("Invalid IBAN")
    NO_BIC = _("No BIC")
    INVALID_BIC = _("Invalid BIC")
    BOUNCED = _("Debit bounced")
    RESCINDED = _("Mandate rescinded")
    INACTIVE = _("Direct debit deactivated")
    NO_MANDATE_REFERENCE = _("No mandate reference")


class MemberSepa(Auditable, models.Model):
    member = AutoOneToOneField(
        to="members.Member", related_name="profile_sepa", on_delete=models.PROTECT
    )

    iban = IBANField(null=True, blank=True, verbose_name="IBAN")
    bic = BICField(null=True, blank=True, verbose_name="BIC")

    mandate_state = models.CharField(
        choices=[
            ("inactive", _("Inactive")),
            ("active", _("Active")),
            ("bounced", _("Bounced")),
            ("rescinded", _("Rescinded")),
        ],
        default="active",
        max_length=10,
        blank=False,
        null=False,
        verbose_name=_("Mandate state"),
    )

    institute = models.CharField(
        max_length=255, null=True, blank=True, verbose_name=_("IBAN Institute")
    )

    issue_date = models.DateField(
        null=True,
        blank=True,
        verbose_name=_("IBAN Issue Date"),
        help_text=_(
            "The issue date of the direct debit mandate. (1970-01-01 means there is no issue date in the database )"
        ),
    )

    fullname = models.CharField(
        null=True,
        blank=True,
        max_length=255,
        verbose_name=_("IBAN full name"),
        help_text=_("Full name for IBAN account owner"),
    )

    address = models.CharField(
        null=True,
        blank=True,
        max_length=255,
        verbose_name=_("IBAN address"),
        help_text=_("Address line (e.g. Street / House Number)"),
    )

    zip_code = models.CharField(
        null=True,
        blank=True,
        max_length=20,
        verbose_name=_("IBAN zip code"),
        help_text=_("ZIP Code"),
    )

    city = models.CharField(
        null=True, blank=True, max_length=255, verbose_name=_("IBAN City")
    )

    country = models.CharField(
        null=True,
        blank=True,
        max_length=255,
        default="Deutschland",
        verbose_name=_("IBAN Country"),
    )

    mandate_reference = models.CharField(
        null=True, blank=True, max_length=255, verbose_name=_("IBAN Mandate Reference")
    )

    mandate_reason = models.CharField(
        null=True, blank=True, max_length=255, verbose_name=_("IBAN Mandate Reason")
    )

    form_title = _("SEPA information")

    @property
    def is_usable(self):
        return self.sepa_direct_debit_state == SepaDirectDebitState.OK

    @property
    def iban_parsed(self):
        with suppress(ValueError):
            if self.iban:
                return IBAN(self.iban)

        return None

    @property
    def bic_autocomplete(self):
        if self.bic:
            return self.bic

        iban_parsed = self.iban_parsed
        if not iban_parsed:
            return None

        with suppress(ValueError):
            return str(iban_parsed.bic)

        return None

    @property
    def bic_parsed(self):
        with suppress(ValueError):
            bic = self.bic_autocomplete
            if bic:
                return BIC(bic)

        return None

    @property
    def sepa_direct_debit_state(self):
        if not self.iban:
            return SepaDirectDebitState.NO_IBAN

        if not self.iban_parsed:
            return SepaDirectDebitState.INVALID_IBAN

        if self.mandate_state == "rescinded":
            return SepaDirectDebitState.RESCINDED

        if self.mandate_state == "bounced":
            return SepaDirectDebitState.BOUNCED

        if self.mandate_state == "inactive":
            return SepaDirectDebitState.INACTIVE

        if not self.bic_autocomplete:
            return SepaDirectDebitState.NO_BIC

        bic = self.bic_parsed
        if not bic:
            return SepaDirectDebitState.INVALID_BIC

        if bic.country_code == "DE" and not bic.exists:
            # PBNKDEFF and PBNKDEFFXXX should be the same
            b_ = BIC(str(bic) + "XXX")
            if not b_.exists:
                return SepaDirectDebitState.INVALID_BIC

        if not self.mandate_reference:
            return SepaDirectDebitState.NO_MANDATE_REFERENCE

        return SepaDirectDebitState.OK



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class BookkeepingConfig(AppConfig):
    name = "byro.bookkeeping"

    class ByroPluginMeta:
        document_categories = {
            "byro.bookkeeping.receipt": _("Receipt"),
            "byro.bookkeeping.invoice": _("Invoice"),
            "byro.bookkeeping.account.statement": _("Statement"),
        }



from collections import OrderedDict
from decimal import Decimal

from django import forms
from django.db import models
from django.utils.functional import classproperty
from django.utils.translation import gettext_lazy as _

from byro.common.models import Configuration
from byro.members.models import Member, Membership


class DefaultDates:
    TODAY = "today"
    BEGINNING_MONTH = "beginning_month"
    BEGINNING_MONTH_NEXT = "beginning_month_next"
    BEGINNING_YEAR = "beginning_year"
    BEGINNING_YEAR_NEXT = "beginning_year_next"
    FIXED_DATE = "fixed_date"

    @classproperty
    def choices(cls):
        return (
            (None, "------------"),
            (cls.TODAY, _("Current day")),
            (cls.BEGINNING_MONTH, _("Beginning of current month")),
            (cls.BEGINNING_MONTH_NEXT, _("Beginning of next month")),
            (cls.BEGINNING_YEAR, _("Beginning of current year")),
            (cls.BEGINNING_YEAR_NEXT, _("Beginning of next year")),
            (cls.FIXED_DATE, _("Other/fixed date")),
        )


class DefaultBoolean:
    @classproperty
    def choices(cls):
        return ((None, "------------"), (False, _("False")), (True, _("True")))


SPECIAL_NAMES = {Member: "member", Membership: "membership"}
SPECIAL_ORDER = [
    "member__number",
    "member__name",
    "member__address",
    "member__email",
    "membership__start",
    "membership__interval",
    "membership__amount",
]


class RegistrationConfigForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields_extra = OrderedDict()
        fieldsets = []
        config = Configuration.get_solo().registration_form or []
        data = {entry["name"]: entry for entry in config if "name" in entry}

        for model, field in self.get_form_fields():
            key = f"{SPECIAL_NAMES.get(model, model.__name__)}__{field.name}"
            entry = data.get(key, {})

            verbose_name = field.verbose_name or field.name
            if model not in SPECIAL_NAMES:
                verbose_name = "{verbose_name} ({model.__name__})".format(
                    verbose_name=verbose_name, model=model
                )

            fields = OrderedDict()
            fields["position"] = forms.IntegerField(
                required=False, label=_("Position in form")
            )
            if isinstance(field, models.DateField):
                fields["default_date"] = forms.ChoiceField(
                    required=False,
                    label=_("Default date"),
                    choices=DefaultDates.choices,
                )
            if isinstance(field, models.BooleanField):
                fields["default_boolean"] = forms.ChoiceField(
                    required=False,
                    label=_("Default value"),
                    choices=DefaultBoolean.choices,
                )
            default_field = self.build_default_field(field, model)
            if default_field:
                fields["default"] = default_field
            for name, form_field in fields.items():
                form_field.initial = entry.get(name, form_field.initial)

            fieldsets.append(
                (
                    (  # This part is responsible for sorting the model fields:
                        data.get(key, {}).get("position", None)
                        or 998,  # Position in form, if set (or 998)
                        SPECIAL_ORDER.index(key)
                        if key in SPECIAL_ORDER
                        else 66,  # SPECIAL_ORDER first
                        0 if model in SPECIAL_NAMES else 1,  # SPECIAL_NAMES first
                    ),
                    key,  # Fall back to sorting by key, otherwise
                    verbose_name,
                    OrderedDict(
                        (
                            f"{key}__{name}",
                            value,
                        )  # TODO: make fields an ordered dict that prepends {key} to every key for more fanciness
                        for name, value in fields.items()
                    ),
                )
            )

        fieldsets.sort()
        for _position, key, verbose_name, form_fields in fieldsets:
            self.fields_extra[key] = (
                verbose_name,
                (self[name] for name in form_fields.keys()),
            )
            self.fields.update(form_fields)

    @staticmethod
    def get_form_fields():
        for model in [Member, Membership] + Member.profile_classes:
            for field in model._meta.fields:
                if field.name in ("id", "member") or (
                    model is Member
                    and field.name
                    in ["membership_type", "direct_address_name", "order_name"]
                ):
                    continue
                yield (model, field)

    def build_default_field(self, field, model):
        choices = getattr(field, "choices", None)
        if choices:
            return forms.ChoiceField(
                required=False,
                label=_("Default value"),
                choices=[(None, "-----------")] + list(choices),
            )
        if not (model is Member and field.name == "number"):
            if isinstance(field, models.CharField):
                return forms.CharField(required=False, label=_("Default value"))
            elif isinstance(field, models.DecimalField):
                return forms.DecimalField(
                    required=False,
                    label=_("Default value"),
                    max_digits=field.max_digits,
                    decimal_places=field.decimal_places,
                )
            elif isinstance(field, models.DateField):
                return forms.CharField(required=False, label=_("Other/fixed date"))

    def clean(self):
        ret = super().clean()
        positions = [
            value
            for (key, value) in ret.items()
            if key.endswith("__position") and value is not None
        ]
        if not len(list(positions)) == len(set(positions)):
            raise forms.ValidationError("Every position must be unique!")
        return ret

    def save(self):
        data = {}
        for full_name, value in self.cleaned_data.items():
            name, key = full_name.rsplit("__", 1)
            if not (value == "" or value is None):
                if isinstance(value, Decimal):
                    value = str(value)
                if key == "default_boolean":
                    value = bool(value == "True")
                data.setdefault(name, {})[key] = value
        data = [dict(name=key, **value) for (key, value) in data.items()]
        config = Configuration.get_solo()
        config.registration_form = list(data)
        config.save()



from annoying.fields import AutoOneToOneField
from django.db import models
from django.utils.translation import gettext_lazy as _

from byro.common.models.auditable import Auditable


class MemberProfile(Auditable, models.Model):
    member = AutoOneToOneField(
        to="members.Member", related_name="profile_profile", on_delete=models.PROTECT
    )
    nick = models.CharField(
        verbose_name=_("Nick"), max_length=200, blank=True, null=True
    )
    birth_date = models.DateField(verbose_name=_("Birth date"), blank=True, null=True)
    phone_number = models.CharField(
        verbose_name=_("Phone number"), max_length=32, blank=True, null=True
    )

    form_title = _("General information")



import configparser
import os
import sys

from byro.common.settings.utils import reduce_dict

CONFIG = {
    "filesystem": {
        "base": {
            "default": os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
            )
        },
        "logs": {"default": None, "env": os.getenv("BYRO_FILESYSTEM_LOGS")},
        "media": {"default": None, "env": os.getenv("BYRO_FILESYSTEM_MEDIA")},
        "static": {"default": None, "env": os.getenv("BYRO_FILESYSTEM_STATIC")},
    },
    "site": {
        "debug": {"default": "runserver" in sys.argv, "env": os.getenv("BYRO_DEBUG")},
        "url": {"default": "http://localhost", "env": os.getenv("BYRO_SITE_URL")},
        "https": {"env": os.getenv("BYRO_HTTPS")},
    },
    "database": {
        "name": {"env": os.getenv("BYRO_DB_NAME")},
        "user": {"default": "", "env": os.getenv("BYRO_DB_USER")},
        "password": {"default": "", "env": os.getenv("BYRO_DB_PASS")},
        "host": {"default": "", "env": os.getenv("BYRO_DB_HOST")},
        "port": {"default": "", "env": os.getenv("BYRO_DB_PORT")},
        "engine": {"default": "postgresql", "env": os.getenv("BYRO_DB_ENGINE")},
    },
    "mail": {
        "from": {"default": "admin@localhost", "env": os.getenv("BYRO_MAIL_FROM")},
        "host": {"default": "localhost", "env": os.getenv("BYRO_MAIL_HOST")},
        "port": {"default": "25", "env": os.getenv("BYRO_MAIL_PORT")},
        "user": {"default": "", "env": os.getenv("BYRO_MAIL_USER")},
        "password": {"default": "", "env": os.getenv("BYRO_MAIL_PASSWORD")},
        "tls": {"default": "False", "env": os.getenv("BYRO_MAIL_TLS")},
        "ssl": {"default": "False", "env": os.getenv("BYRO_MAIL_SSL")},
    },
    "logging": {
        "email": {"default": "", "env": os.getenv("BYRO_LOGGING_EMAIL")},
        "email_level": {"default": "", "env": os.getenv("BYRO_LOGGING_EMAIL_LEVEL")},
    },
    "locale": {
        "language_code": {"default": "en", "env": os.getenv("BYRO_LANGUAGE_CODE")},
        "time_zone": {"default": "UTC", "env": os.getenv("BYRO_TIME_ZONE")},
    },
}


def read_config_files(config):
    if "BYRO_CONFIG_FILE" in os.environ:
        config_files = config.read_file(
            open(os.environ.get("BYRO_CONFIG_FILE"), encoding="utf-8")
        )
    else:
        config_files = config.read(
            ["/etc/byro/byro.cfg", os.path.expanduser("~/.byro.cfg"), "byro.cfg"],
            encoding="utf-8",
        )
    return (
        config,
        config_files or [],
    )  # .read() returns None if there are no config files


def read_layer(layer_name, config):
    config_dict = reduce_dict(
        {
            section_name: {
                key: value.get(layer_name) for key, value in section_content.items()
            }
            for section_name, section_content in CONFIG.items()
        }
    )
    config.read_dict(config_dict)
    return config


def build_config():
    config = configparser.RawConfigParser()
    config = read_layer("default", config)
    config, config_files = read_config_files(config)
    config = read_layer("env", config)
    return config, config_files



# Generated by Django 1.11.8 on 2018-02-24 21:14

from django.db import migrations


def init_templates(apps, schema_editor):
    from byro.mails import default

    MailTemplate = apps.get_model("mails", "MailTemplate")
    Configuration = apps.get_model("common", "Configuration")
    config, _ = Configuration.objects.get_or_create()
    if not config.welcome_member_template:
        welcome_member = MailTemplate.objects.create(
            subject=default.WELCOME_MEMBER_SUBJECT, text=default.WELCOME_MEMBER_TEXT
        )
        config.welcome_member_template = welcome_member
    if not config.welcome_office_template:
        welcome_office = MailTemplate.objects.create(
            subject=default.WELCOME_OFFICE_SUBJECT, text=default.WELCOME_OFFICE_TEXT
        )
        config.welcome_office_template = welcome_office
    config.save()


class Migration(migrations.Migration):
    dependencies = [("common", "0006_auto_20180224_2114")]

    operations = [migrations.RunPython(init_templates, migrations.RunPython.noop)]



from django import forms
from django.contrib import messages
from django.db import transaction
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView
from django_select2.forms import Select2Widget

from byro.bookkeeping.models import (
    Account,
    Booking,
    DocumentTransactionLink,
    Transaction,
)

from .documents import DocumentUploadForm


class NewBookingForm(forms.Form):
    memo = forms.CharField(label=_("Memo"), max_length=1000, required=False)
    member = Booking._meta.get_field("member").formfield(widget=Select2Widget)
    account = Booking._meta.get_field("debit_account").formfield(widget=Select2Widget)
    debit_value = forms.DecimalField(
        min_value=0, max_digits=8, decimal_places=2, required=False
    )
    credit_value = forms.DecimalField(
        min_value=0, max_digits=8, decimal_places=2, required=False
    )


class TransactionDetailView(ListView):
    template_name = "office/transaction/detail.html"
    context_object_name = "bookings"
    model = Transaction
    paginate_by = None

    def get_form(self, input_data=None):
        form = NewBookingForm(input_data)
        form.fields["account"].required = True
        form.fields["member"].required = False
        t = self.get_object()
        if t.balances_credit != t.balances_debit:
            if t.balances_credit > t.balances_debit:
                form.fields["debit_value"].initial = (
                    t.balances_credit - t.balances_debit
                )
            else:
                form.fields["credit_value"].initial = (
                    t.balances_debit - t.balances_credit
                )
        return form

    def get_upload_form(self, input_data=None, input_files=None):
        form = DocumentUploadForm(
            input_data,
            input_files,
            prefix="upload_form",
            initial_category="byro.bookkeeping.receipt",
        )
        form.fields["title"].required = False
        return form

    def get_object(self):
        return Transaction.objects.with_balances().get(pk=self.kwargs["pk"])

    def get_queryset(self):
        return self.get_object().bookings.all()

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["transaction"] = self.get_object()
        context["form"] = self.get_form()
        context["upload_form"] = self.get_upload_form()
        return context

    @transaction.atomic
    def post(self, request, *args, **kwargs):
        t = self.get_object()
        form = self.get_form(request.POST)
        upload_form = self.get_upload_form(request.POST, request.FILES)

        if upload_form.is_valid():
            self.process_upload_form(upload_form, t)

        if form.is_valid():
            self.process_transaction_changes(form, t)
            t = self.get_object()

            if t.is_balanced:
                if "in_account" in request.GET:
                    account = Account.objects.get(pk=request.GET["in_account"])
                    if account.unbalanced_transactions.count():
                        return redirect(
                            "{}?filter=unbalanced".format(
                                reverse(
                                    "office:finance.accounts.detail",
                                    kwargs={"pk": account.pk},
                                )
                            )
                        )
                return redirect("office:finance.accounts.list")

        if "in_account" in request.GET:
            return redirect(
                "{}?in_account={}".format(
                    reverse("office:finance.transactions.detail", kwargs={"pk": t.pk}),
                    request.GET["in_account"],
                )
            )
        else:
            return redirect("office:finance.transactions.detail", pk=t.pk)

    def process_transaction_changes(self, form, t):
        arguments = dict(
            memo=form.cleaned_data["memo"],
            account=form.cleaned_data["account"],
            member=form.cleaned_data["member"],
            importer="_manual_entry",
            user_or_context=self,
        )
        if form.cleaned_data["debit_value"]:
            t.debit(amount=form.cleaned_data["debit_value"], **arguments)
        if form.cleaned_data["credit_value"]:
            t.credit(amount=form.cleaned_data["credit_value"], **arguments)
        t.save()
        messages.success(self.request, _("The transaction was updated."))
        t.log(self, ".updated")

    def process_upload_form(self, form, t):
        form.save()
        DocumentTransactionLink.objects.create(transaction=t, document=form.instance)
        t.log(
            self,
            ".document.created",
            document=form.instance,
            content_hash=form.instance.content_hash,
        )



from django.utils.translation import gettext_lazy as _
from i18nfield.strings import LazyI18nString

WELCOME_MEMBER_SUBJECT = LazyI18nString.from_gettext(_("Welcome, latest member!"))
WELCOME_MEMBER_TEXT = LazyI18nString.from_gettext(
    _(
        """Hi,

welcome to {name}! You're now officially our latest member. Your member ID
is {number}. If you have any questions relating to your member fees or
you want to update your member data, please contact us at {contact}.

{additional_information}

Thanks,
the robo clerk"""
    )
)

WELCOME_OFFICE_SUBJECT = LazyI18nString.from_gettext(_("[byro] New member"))
WELCOME_OFFICE_TEXT = LazyI18nString.from_gettext(
    _(
        """Hi,

we have a new member: {member_name}

{additional_information}

Thanks,
the robo clerk"""
    )
)

LEAVE_MEMBER_SUBJECT = LazyI18nString.from_gettext(_("Goodbye!"))
LEAVE_MEMBER_TEXT = LazyI18nString.from_gettext(
    _(
        """Hi,

we are sorry that you will leave us at {end}.

{additional_information}

Thanks,
the robo clerk"""
    )
)

LEAVE_MEMBER_OFFICE_SUBJECT = LazyI18nString.from_gettext(
    _("[byro] Membership termination")
)
LEAVE_MEMBER_OFFICE_TEXT = LazyI18nString.from_gettext(
    _(
        """Hi,

{member_name} will leave us at {end}.

{additional_information}

Thanks,
the robo clerk"""
    )
)
RECORD_DISCLOSURE_SUBJECT = LazyI18nString.from_gettext(
    _("Your {association_name} data record (#{number})")
)
RECORD_DISCLOSURE_TEXT = LazyI18nString.from_gettext(
    _(
        """Hi,

We're writing you to let you know about the data we have currently stored about
yourself and your membership.

Your current balance is {balance}.

{data}

If these records are outdated or incorrect, please let us know as a response to
this email.

Thanks,
the robo clerk"""
    )
)



from django import forms
from django.utils.translation import gettext_lazy as _


class PrivacyConsentForm(forms.Form):
    is_visible_to_members = forms.BooleanField(
        label=_(
            "Yes, my data may be shown in the member list, where it will be visible to other members."
        ),
        required=False,
    )

    def __init__(self, *args, member=None, **kwargs):
        initial = kwargs.pop("initial", {})
        initial[
            "is_visible_to_members"
        ] = member.profile_memberpage.is_visible_to_members
        super().__init__(*args, initial=initial, **kwargs)
        self.member = member

        blocked = [
            "sepa",
            "secret",
            "balance",
            "active",
            "_internal_id",
            "memberpage",
            "memberships.last",
        ]

        for key, value in member.get_fields().items():
            if (
                any(b in value.path for b in blocked)
                or any(key in b for b in blocked)
                or key == "pk"
            ):
                continue
            self.fields[key] = forms.BooleanField(
                required=False,
                label=value.name,
                initial=self.member.profile_memberpage.publication_consent.get(
                    "fields", {}
                )
                .get(key, {})
                .get("visibility"),
            )

    def save(self):
        data = self.cleaned_data
        result = {}
        for key in self.member.get_fields():
            if data.get(key):
                result[key] = {"visibility": "share"}
        self.member.profile_memberpage.publication_consent = {"fields": result}
        self.member.profile_memberpage.is_visible_to_members = data[
            "is_visible_to_members"
        ]
        self.member.profile_memberpage.save()



# Generated by Django 1.11.8 on 2018-02-26 16:36

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("documents", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="document",
            name="title",
            field=models.CharField(max_length=300, null=True),
        )
    ]



from django import forms
from django.apps import apps
from django.contrib import messages
from django.db import transaction
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import FormView, TemplateView, ListView

from byro.bookkeeping.models import Account
from byro.common.forms import ConfigurationForm, RegistrationConfigForm, InitialForm
from byro.common.models import LogEntry
from byro.common.models.configuration import Configuration, ByroConfiguration


class InitialSettings(FormView):
    form_class = InitialForm
    template_name = "office/settings/initial.html"

    def get_form_kwargs(self):
        form_kwargs = super().get_form_kwargs()
        form_kwargs["instance"] = Configuration.get_solo()
        return form_kwargs

    @transaction.atomic
    def form_valid(self, form):
        form.save()
        other_data = {
            "accounts": [
                {"account": acc, "name": acc.name, "balances": acc.balances()}
                for acc in Account.objects.all()
            ]
        }
        form.instance.log(
            self, ".initial", initial_data=form.cleaned_data, other_data=other_data
        )
        messages.success(
            self.request,
            _(
                "You're nearly ready to go – configure how you want to add new members, and you're done."
            ),
        )
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("office:settings.registration")


class ConfigurationView(FormView):
    form_class = ConfigurationForm
    template_name = "office/settings/form.html"

    def get_form_kwargs(self):
        form_kwargs = super().get_form_kwargs()
        form_kwargs["instance"] = Configuration.get_solo()
        return form_kwargs

    def get_form(self):
        config_models = [
            model for model in apps.get_models() if issubclass(model, ByroConfiguration)
        ]
        data = self.request.POST if self.request.method == "POST" else None
        return [
            ConfigurationForm(
                prefix=Configuration.__name__,
                instance=Configuration.get_solo(),
                data=data,
            )
        ] + [
            forms.modelform_factory(model, fields="__all__")(
                prefix=model.__name__, instance=model.get_solo(), data=data
            )
            for model in config_models
            if not issubclass(model, Configuration)
        ]

    @transaction.atomic
    def form_valid(self, form):
        for f in form:
            f.save()
            if f.changed_data:
                f.instance.log(
                    self,
                    "byro.settings.changed",
                    changes={
                        k: (self._original_values[f][k], f.cleaned_data[k])
                        for k in f.changed_data
                    },
                )

        messages.success(self.request, _("The config was saved successfully."))
        return super().form_valid(f)

    def post(self, request, *args, **kwargs):
        form = self.get_form()
        self._original_values = {
            f: {k: getattr(f.instance, k, None) for k in f.fields} for f in form
        }
        if all(f.is_valid() for f in form):
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def get_success_url(self):
        return reverse("office:settings.base")


class RegistrationConfigView(FormView):
    form_class = RegistrationConfigForm
    template_name = "office/settings/registration_form.html"

    def form_valid(self, form):
        form.save()
        messages.success(self.request, _("The config was saved successfully."))
        LogEntry.objects.create(
            content_object=Configuration.get_solo(),
            user=self.request.user,
            action_type="byro.settings.registration.changed",
        )
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("office:settings.registration")


class AboutByroView(TemplateView):
    template_name = "office/settings/about.html"

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["plugins"] = []
        for app in apps.get_app_configs():
            if hasattr(app, "ByroPluginMeta") and hasattr(app.ByroPluginMeta, "name"):
                context["plugins"].append({"meta": app.ByroPluginMeta})
        return context


class LogView(ListView):
    template_name = "office/settings/log.html"
    context_object_name = "log_entries"
    model = LogEntry
    paginate_by = 50



# Generated by Django 1.11.4 on 2017-08-12 11:03

import byro.common.models.auditable
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [("members", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="Account",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "account_category",
                    models.CharField(
                        choices=[
                            ("member_donation", "member_donation"),
                            ("member_fees", "member_fees"),
                        ],
                        max_length=15,
                    ),
                ),
                ("name", models.CharField(max_length=300)),
                (
                    "member",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="members.Member",
                    ),
                ),
            ],
            bases=(byro.common.models.auditable.Auditable, models.Model),
        ),
        migrations.CreateModel(
            name="RealTransaction",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "channel",
                    models.CharField(
                        choices=[("bank", "bank"), ("cash", "cash")], max_length=4
                    ),
                ),
                ("booking_datetime", models.DateTimeField(null=True)),
                ("value_datetime", models.DateTimeField()),
                ("amount", models.DecimalField(decimal_places=2, max_digits=8)),
                ("purpose", models.CharField(max_length=200)),
                ("originator", models.CharField(max_length=200)),
                ("importer", models.CharField(max_length=200, null=True)),
                ("data", django.contrib.postgres.fields.jsonb.JSONField(null=True)),
                (
                    "reverses",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="bookkeeping.RealTransaction",
                    ),
                ),
            ],
            bases=(byro.common.models.auditable.Auditable, models.Model),
        ),
        migrations.CreateModel(
            name="VirtualTransaction",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("amount", models.DecimalField(decimal_places=2, max_digits=8)),
                ("value_datetime", models.DateTimeField(null=True)),
                (
                    "destination_account",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="incoming_transactions",
                        to="bookkeeping.Account",
                    ),
                ),
                (
                    "real_transaction",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="virtual_transactions",
                        to="bookkeeping.RealTransaction",
                    ),
                ),
                (
                    "source_account",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="outgoing_transactions",
                        to="bookkeeping.Account",
                    ),
                ),
            ],
            bases=(byro.common.models.auditable.Auditable, models.Model),
        ),
    ]



# Generated by Django 2.2.12 on 2020-04-25 09:42

import byro.public.models
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("public", "0002_memberpageprofile_publication_consent"),
    ]

    operations = [
        migrations.AlterField(
            model_name="memberpageprofile",
            name="secret_token",
            field=models.CharField(
                blank=True,
                default=byro.public.models.generate_default_token,
                max_length=128,
                null=True,
                unique=True,
            ),
        ),
    ]



# Generated by Django 1.11.8 on 2018-01-13 13:12

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("bookkeeping", "0008_auto_20180111_1807")]

    operations = [
        migrations.AddField(
            model_name="realtransactionsource",
            name="state",
            field=models.CharField(
                choices=[
                    ("new", "new"),
                    ("processing", "processing"),
                    ("processed", "processed"),
                    ("failed", "failed"),
                ],
                default="new",
                max_length=10,
            ),
        )
    ]



from django import template

from byro.common.models import Configuration

register = template.Library()


@register.filter
def format_with_currency(value, long_form=False):
    config = Configuration.get_solo()
    amount = str(value if config.display_cents else int(value))
    currency = config.currency if long_form else config.currency_symbol
    if currency:
        parts = [amount, currency]
        return " ".join(parts if config.currency_postfix else reversed(parts))
    else:
        return amount



# Generated by Django 2.2.12 on 2020-08-09 19:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("members", "0010_memberbalance"),
    ]

    operations = [
        migrations.AddField(
            model_name="member",
            name="direct_address_name",
            field=models.CharField(
                blank=True,
                max_length=100,
                null=True,
                verbose_name="Name for direct address",
            ),
        ),
        migrations.AddField(
            model_name="member",
            name="order_name",
            field=models.CharField(
                blank=True, max_length=100, null=True, verbose_name="Name (sort order)"
            ),
        ),
    ]



# Generated by Django 1.11.13 on 2018-06-07 20:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("profile", "0004_auto_20171206_1919")]

    operations = [
        migrations.AlterField(
            model_name="memberprofile",
            name="birth_date",
            field=models.DateField(blank=True, null=True, verbose_name="Birth date"),
        ),
        migrations.AlterField(
            model_name="memberprofile",
            name="nick",
            field=models.CharField(
                blank=True, max_length=200, null=True, verbose_name="Nick"
            ),
        ),
    ]



import subprocess
import sys
from contextlib import suppress

from django.apps import apps


def get_plugins():
    result = []
    for app in apps.get_app_configs():
        if hasattr(app, "ByroPluginMeta"):
            if getattr(app, "name", "").startswith("byro.") and not hasattr(
                app.ByroPluginMeta, "version"
            ):
                continue
            result.append(app)
    return result


def get_version():
    with suppress(Exception):
        retval = getattr(sys.modules[__name__], "_byro_git_version", None)
        if retval:
            return retval

        retval = (
            subprocess.check_output(
                ["git", "describe", "--always", "--dirty", "--abbrev=40"],
                stderr=subprocess.PIPE,
            )
            .decode()
            .strip()
        )
        sys.modules[__name__]._byro_git_version = retval
        return retval

    with suppress(Exception):
        from byro import __version__

        return __version__

    return ""


def get_installed_software():
    retval = [f"byro {get_version()}"]
    for plugin in get_plugins():
        retval.append(
            "{} {}".format(
                plugin.name, getattr(plugin.ByroPluginMeta, "version", "")
            ).strip()
        )
    return retval



"""byro URL Configuration.

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.11/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.conf.urls import url, include
    2. Add a URL to urlpatterns:  url(r'^blog/', include('blog.urls'))
"""
import importlib.util
from contextlib import suppress

from django.apps import apps
from django.conf import settings
from django.conf.urls.static import static
from django.urls import include, re_path

raw_plugin_patterns = []
for app in apps.get_app_configs():
    if hasattr(app, "ByroPluginMeta"):
        if importlib.util.find_spec(app.name + ".urls"):
            urlmod = importlib.import_module(app.name + ".urls")
            single_plugin_patterns = []
            if hasattr(urlmod, "urlpatterns"):
                single_plugin_patterns += urlmod.urlpatterns
            raw_plugin_patterns.append(
                re_path(r"", include((single_plugin_patterns, app.label)))
            )

urlpatterns = [
    re_path(r"", include((raw_plugin_patterns, "plugins"))),
    re_path(r"", include("byro.common.urls", namespace="common")),
    re_path(r"", include("byro.office.urls", namespace="office")),
    re_path(r"^p/", include("byro.public.urls", namespace="public")),
]

if settings.DEBUG:
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

    with suppress(ImportError):
        import debug_toolbar

        urlpatterns.insert(0, re_path(r"^__debug__/", include(debug_toolbar.urls)))



# Generated by Django 1.11.6 on 2017-10-12 20:32

import annoying.fields
from django.db import migrations
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [("profile", "0002_auto_20171012_1915")]

    operations = [
        migrations.AlterField(
            model_name="memberprofile",
            name="member",
            field=annoying.fields.AutoOneToOneField(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="profile_profile",
                to="members.Member",
            ),
        )
    ]



# Generated by Django 1.11.8 on 2018-01-24 14:39

import byro.common.models.auditable
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [("members", "0006_auto_20180113_1849")]

    operations = [
        migrations.CreateModel(
            name="Document",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("document", models.FileField(upload_to="documents/")),
                ("category", models.CharField(max_length=300, null=True)),
                (
                    "direction",
                    models.CharField(
                        choices=[("incoming", "incoming"), ("outgoing", "outgoing")],
                        default="outgoing",
                        max_length=8,
                    ),
                ),
                (
                    "member",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="documents",
                        to="members.Member",
                    ),
                ),
            ],
            bases=(byro.common.models.auditable.Auditable, models.Model),
        )
    ]



# Generated by Django 1.11.4 on 2017-08-12 11:09

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("bookkeeping", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="name",
            field=models.CharField(max_length=300, null=True),
        )
    ]



# Generated by Django 2.2.12 on 2020-07-30 23:57

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("common", "0015_auto_20200421_2336"),
    ]

    operations = [
        migrations.AddField(
            model_name="configuration",
            name="accounting_start",
            field=models.DateField(
                blank=True,
                help_text="This field is especially useful if the organization was later migrated to byro and the membership fees of members from the past are not to be billed. Leave the field empty if you do not have this requirement and you want to invoice all members from the beginning of their membership.",
                null=True,
                verbose_name="Start accounting Membership Fees from",
            ),
        ),
    ]



# Generated by Django 1.11.9 on 2018-01-13 18:49

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("members", "0005_auto_20171206_1919")]

    operations = [
        migrations.AlterField(
            model_name="member",
            name="number",
            field=models.CharField(
                blank=True,
                db_index=True,
                max_length=100,
                null=True,
                verbose_name="Membership number/ID",
            ),
        )
    ]



from .accounts import (
    AccountCreateView,
    AccountDeleteView,
    AccountDetailView,
    AccountListView,
)
from .dashboard import DashboardView
from .members import (
    MemberCreateView,
    MemberDashboardView,
    MemberDataView,
    MemberFinanceView,
    MemberListExportView,
    MemberListImportView,
    MemberListView,
    MemberMailsView,
    MemberOperationsView,
    MemberTimelineView,
)
from .settings import ConfigurationView, RegistrationConfigView
from .transactions import TransactionDetailView
from .users import UserCreateView, UserDetailView, UserListView

__all__ = [
    "AccountCreateView",
    "AccountDeleteView",
    "AccountDetailView",
    "AccountListView",
    "ConfigurationView",
    "DashboardView",
    "MemberCreateView",
    "MemberDashboardView",
    "MemberDataView",
    "MemberFinanceView",
    "MemberOperationsView",
    "MemberListView",
    "MemberListExportView",
    "MemberListImportView",
    "MemberMailsView",
    "MemberTimelineView",
    "UserListView",
    "UserCreateView",
    "UserDetailView",
    "RegistrationConfigView",
    "TransactionDetailView",
]



from django.urls import re_path

from .views import LogInfoView, LoginView, logout_view

app_name = "common"
urlpatterns = [
    re_path("^login/$", LoginView.as_view(), name="login"),
    re_path("^logout/$", logout_view, name="logout"),
    re_path("^log/info$", LogInfoView.as_view(), name="log.info"),
]



# Generated by Django 2.1.8 on 2019-04-15 12:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [("members", "0009_auto_20180512_1810")]

    operations = [
        migrations.CreateModel(
            name="MemberBalance",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "reference",
                    models.CharField(
                        blank=True,
                        help_text="For example an invoice number or a payment reference",
                        max_length=50,
                        null=True,
                        unique=True,
                        verbose_name="Reference",
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2, max_digits=8, verbose_name="Amount"
                    ),
                ),
                ("start", models.DateTimeField(verbose_name="Start")),
                ("end", models.DateTimeField(verbose_name="End")),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("paid", "paid"),
                            ("partial", "partially paid"),
                            ("unpaid", "unpaid"),
                        ],
                        default="unpaid",
                        max_length=7,
                    ),
                ),
                (
                    "member",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="balances",
                        to="members.Member",
                    ),
                ),
            ],
        )
    ]



# Generated by Django 2.0 on 2017-12-06 19:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("bookkeeping", "0006_remove_account_member")]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="account_category",
            field=models.CharField(
                choices=[
                    ("member_donation", "Donation account"),
                    ("member_fees", "Membership fee account"),
                    ("asset", "Asset account"),
                    ("liability", "Liability account"),
                    ("income", "Income account"),
                    ("expense", "Expense account"),
                ],
                max_length=15,
            ),
        )
    ]



from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _

from byro.members.signals import (
    leave_member_office_mail_information,
    new_member_mail_information,
    new_member_office_mail_information,
)


@receiver(new_member_mail_information)
def new_member_mail_info_sepa(sender, signal, **kwargs):
    if not (
        hasattr(sender, "profile_sepa")
        and sender.profile_sepa
        and sender.profile_sepa.is_usable
    ):
        return ""
    return _("Your SEPA mandate reference is {}.").format(
        sender.profile_sepa.mandate_reference
    )


@receiver(new_member_office_mail_information)
def new_member_office_mail_info_sepa(sender, signal, **kwargs):
    if not (
        hasattr(sender, "profile_sepa")
        and sender.profile_sepa
        and sender.profile_sepa.is_usable
    ):
        return ""
    if not sender.memberships.count() == 1:
        raise Exception("Cannot determine which membership to put in email!")
    membership = sender.memberships.first()
    data = {
        "iban": sender.profile_sepa.iban,
        "bic": sender.profile_sepa.bic,
        "institute": sender.profile_sepa.institute,
        "issue_date": sender.profile_sepa.issue_date.isoformat(),
        "name": sender.profile_sepa.fullname,
        "mandate_reference": sender.profile_sepa.mandate_reference,
        "mandate_reason": sender.profile_sepa.mandate_reason,
        "amount": f"{membership.amount:2f}",
        "interval": membership.get_interval_display(),
        "start": membership.start.isoformat(),
    }
    return _(
        """The new member has given us a SEPA mandate for {amount} ({interval}), starting on {start}:

Name: {name}
IBAN: {iban}
BIC: {bic} ({institute})
Mandate date: {issue_date}
Mandate reference: {mandate_reference}
Mandate reason: {mandate_reason}

"""
    ).format(**data)


@receiver(leave_member_office_mail_information)
def leave_member_office_mail_info_sepa(sender, signal, **kwargs):
    membership = sender
    member = sender.member

    if not (
        hasattr(member, "profile_sepa")
        and member.profile_sepa
        and member.profile_sepa.is_usable
    ):
        return ""
    data = {
        "number": member.number,
        "iban": member.profile_sepa.iban,
        "bic": member.profile_sepa.bic,
        "institute": member.profile_sepa.institute,
        "issue_date": member.profile_sepa.issue_date.isoformat(),
        "name": member.profile_sepa.fullname,
        "mandate_reference": member.profile_sepa.mandate_reference,
        "mandate_reason": member.profile_sepa.mandate_reason,
        "end": membership.end.isoformat(),
    }
    return _(
        """Please terminate SEPA mandate member number: {number} to {end}:

Name: {name}
IBAN: {iban}
BIC: {bic} ({institute})
Mandate date: {issue_date}
Mandate reference: {mandate_reference}
Mandate reason: {mandate_reason}

"""
    ).format(**data)



# Generated by Django 1.11.6 on 2017-10-12 18:57

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("members", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="member",
            name="address",
            field=models.CharField(blank=True, max_length=300, null=True),
        ),
        migrations.AddField(
            model_name="member",
            name="name",
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AddField(
            model_name="member",
            name="number",
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name="member",
            name="email",
            field=models.EmailField(blank=True, max_length=200, null=True),
        ),
    ]



# Generated by Django 1.11.13 on 2018-07-07 15:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("bookkeeping", "0014_auto_20180707_1410")]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="account_category",
            field=models.CharField(
                choices=[
                    ("asset", "Asset account"),
                    ("liability", "Liability account"),
                    ("income", "Income account"),
                    ("expense", "Expense account"),
                    ("equity", "Equity account"),
                ],
                max_length=9,
            ),
        )
    ]



import django.dispatch

periodic_task = django.dispatch.Signal()
"""
This is a signal that we send out every time the periodic task cronjob runs.
This interval is not sharply defined, it can be everything between a minute and
a day. The actions you perform should be idempotent, i.e. it should not make a
difference if this is sent out more often than expected.
"""

unauthenticated_urls = django.dispatch.Signal()
"""
This signal is used to compile a list of URLs that should bypass the normal
authentication middleware.
The return value must be an iterable where each item is either a) a local
view name, or b) a callable with signature (request, resolver_match) that
should return True if authentication should be bypassed.
Note: in case a) only the local name must be provided, not with the
"plugins:$plugin_name:" namespace prefix, in case b) the callable will see
the request and resolver_match with the full name, including namespace prefix.
"""

log_formatters = django.dispatch.Signal()
"""
This signal is used to compile a list of log entry formatters.
The return value must be a mapping of action_type: callable(LogEntry) -> html_fragment: str
"""



# Generated by Django 1.11.9 on 2018-01-13 15:00

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("bookkeeping", "0009_realtransactionsource_state")]

    operations = [
        migrations.AlterField(
            model_name="realtransaction",
            name="purpose",
            field=models.CharField(max_length=1000),
        )
    ]



from contextlib import suppress

import django.db.utils
from django.db import transaction
from django.utils.functional import classproperty
from django.utils.translation import gettext_lazy as _

from .models import Account, AccountCategory, AccountTag


class SpecialAccounts:
    @classmethod
    def special_account(cls, tag, category, name=None):
        tag, _ignore = AccountTag.objects.get_or_create(name=str(tag).lower())
        accounts = list(
            Account.objects.filter(account_category=category, tags=tag).all()
        )
        if len(accounts) > 1:
            raise Account.MultipleObjectsReturned()
        if accounts:
            account = accounts[0]
        else:
            # Old mechanism: Return an account that is named as the special one would.
            #  But also tag it.
            account = Account.objects.filter(
                account_category=category, name=str(name)
            ).first()
            if not account:
                account = Account.objects.create(
                    account_category=category, name=str(name)
                )
                with suppress(django.db.utils.ProgrammingError):
                    with transaction.atomic():
                        account.log(
                            None,
                            "byro.bookkeeping.account.created",
                            source="Automatic creation of special account",
                        )
            account.tags.add(tag)
            account.save()
        return account

    @classproperty
    def fees(cls):
        return cls.special_account("fees", AccountCategory.INCOME, _("Member fees"))

    @classproperty
    def donations(cls):
        return cls.special_account("donations", AccountCategory.INCOME, _("Donations"))

    @classproperty
    def fees_receivable(cls):
        return cls.special_account(
            "fees_receivable", AccountCategory.ASSET, _("Member fees receivable")
        )

    @classproperty
    def bank(cls):
        return cls.special_account("bank", AccountCategory.ASSET, _("Bank"))

    @classproperty
    def opening_balance(cls):
        return cls.special_account(
            "opening_balance", AccountCategory.ASSET, _("Opening balance")
        )

    @classproperty
    def lost_income(cls):
        return cls.special_account(
            "lost_income", AccountCategory.EXPENSE, _("Lost income")
        )



from codecs import open
from distutils.command.build import build
from os import environ, path

from setuptools import find_packages, setup

from byro import __version__ as byro_version

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
try:
    with open(path.join(here, "../README.rst"), encoding="utf-8") as f:
        long_description = f.read()
except:  # noqa
    long_description = ""


class CustomBuild(build):
    def run(self):
        environ.setdefault("DJANGO_SETTINGS_MODULE", "byro.settings")
        try:
            import django
        except ModuleNotFoundError:
            return
        django.setup()
        from django.conf import settings
        from django.core import management

        settings.COMPRESS_ENABLED = True
        settings.COMPRESS_OFFLINE = True

        management.call_command("compilemessages", verbosity=1)
        management.call_command("collectstatic", verbosity=1, interactive=False)
        management.call_command("compress", verbosity=1)
        build.run(self)


cmdclass = {"build": CustomBuild}


setup(
    name="byro",
    version=byro_version,
    license="Apache License 2.0",
    python_requires=">=3.8",
    description="Membership and fees management for associations, clubs and groups",
    long_description=long_description,
    url="https://byro.cloud",
    author="Tobias Kunze",
    author_email="r@rixx.de",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Intended Audience :: Developers",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    keywords="members membership fees club group clubs associations association",
    install_requires=[
        "canonicaljson==2.0.0",  # https://github.com/matrix-org/python-canonicaljson/blob/master/CHANGES.md
        "celery>=5.1,<6.0",  # search for "what's new" on http://docs.celeryproject.org/en/latest/
        "chardet>=4.0,<5.3",  # https://github.com/chardet/chardet/releases
        "csscompressor~=0.9.5",  # 2017-11, no changelog, https://github.com/sprymix/csscompressor
        "dateparser>=1.1,<1.2",  # https://github.com/scrapinghub/dateparser/blob/master/HISTORY.rst
        "django-annoying~=0.10.0",  # https://github.com/skorokithakis/django-annoying/releases
        "django-bootstrap4>=3.0,<23.3",  # http://django-bootstrap4.readthedocs.io/en/latest/history.html
        "django-compressor>=2.4,<5.0",  # https://django-compressor.readthedocs.io/en/latest/changelog/
        "django-extensions>=3.1,<3.3",  # https://github.com/django-extensions/django-extensions/blob/master/CHANGELOG.md
        "django-formset-js-improved==0.5.0.3",  # no changelog, https://github.com/pretix/django-formset-js
        "django-i18nfield>=1.9,<1.10",  # 2017-11, no changelog, https://github.com/raphaelm/django-i18nfield/
        "django-libsass>=0.8,<0.10",  # inactive, https://github.com/torchbox/django-libsass/blob/master/CHANGELOG.txt
        "django-localflavor>=3.0,<4.1",
        "django-select2>=7.7,<8.2",  # https://github.com/applegrew/django-select2/releases
        "django-solo==2.1.*",  # https://github.com/lazybird/django-solo/blob/master/CHANGES
        "Django>=3.2,<3.3",  # https://docs.djangoproject.com/en/2.0/releases/
        "inlinestyler~=0.2",  # https://github.com/dlanger/inlinestyler/blob/master/CHANGELOG
        "jinja2>=2.10.1",  # https://github.com/pallets/jinja/blob/master/CHANGES.rst
        "more-itertools>=8.10,<10.2",
        "psycopg2-binary",
        "pynacl>=1.4,<1.6",  # https://github.com/pyca/pynacl/blob/master/CHANGELOG.rst
        "python-dateutil",
        "python-magic~=0.4.0",
        "qrcode[pil]>=7.3,<7.5",  # https://github.com/lincolnloop/python-qrcode/blob/master/CHANGES.rst
        "schwifty==2023.6.0",
        "unicodecsv~=0.14.0",
        "whitenoise>=6.4,<6.6",
    ],
    extras_require={
        "dev": [
            "black",
            "check-manifest",
            "djhtml",
            "flake8",
            "flake8-bugbear",
            "freezegun",
            "isort",
            "ipython",
            "pytest",
            "pytest-cov",
            "pytest-django",
            "pytest-rerunfailures",
            "pytest-sugar",
            "pytest-xdist",
        ],
        "mysql": ["mysqlclient"],
        "postgres": ["psycopg2-binary"],
    },
    packages=find_packages(exclude=["tests", "tests.*"]),
    include_package_data=True,
    zip_safe=False,
    cmdclass=cmdclass,
)



# Generated by Django 1.11.8 on 2018-01-24 15:48

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("documents", "0001_initial"), ("mails", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="email",
            name="attachments",
            field=models.ManyToManyField(related_name="mails", to="documents.Document"),
        )
    ]



# Generated by Django 2.1.1 on 2018-09-29 13:17

from django.db import migrations


def init_templates(apps, schema_editor):
    from byro.mails import default

    MailTemplate = apps.get_model("mails", "MailTemplate")
    Configuration = apps.get_model("common", "Configuration")
    config, _ = Configuration.objects.get_or_create()
    if not config.record_disclosure_template:
        template = MailTemplate.objects.create(
            subject=default.RECORD_DISCLOSURE_SUBJECT,
            text=default.RECORD_DISCLOSURE_TEXT,
        )
        config.record_disclosure_template = template
    config.save()


class Migration(migrations.Migration):
    dependencies = [("common", "0011_configuration_record_disclosure_template")]

    operations = [migrations.RunPython(init_templates, migrations.RunPython.noop)]



import logging
from smtplib import SMTPRecipientsRefused, SMTPSenderRefused
from typing import Any, Dict, Union

from django.core.mail import EmailMultiAlternatives, get_connection
from django.core.mail.backends.smtp import EmailBackend
from django.utils.translation import override
from i18nfield.strings import LazyI18nString
from inlinestyler.utils import inline_css

from byro.common.models import Configuration

logger = logging.getLogger(__name__)


class CustomSMTPBackend(EmailBackend):
    def test(self, from_addr):
        try:
            self.open()
            self.connection.ehlo_or_helo_if_needed()
            self.connection.rcpt("test@example.org")
            (code, resp) = self.connection.mail(from_addr, [])
            if code != 250:
                logger.warning(
                    "Error testing mail settings, code %d, resp: %s" % (code, resp)
                )
                raise SMTPSenderRefused(code, resp, from_addr)
            senderrs = {}
            (code, resp) = self.connection.rcpt("test@example.com")
            if (code != 250) and (code != 251):
                logger.warning(
                    "Error testing mail settings, code %d, resp: %s" % (code, resp)
                )
                raise SMTPRecipientsRefused(senderrs)
        finally:
            self.close()


class TolerantDict(dict):
    def __missing__(self, key):
        return key


class SendMailException(Exception):
    pass


def mail(
    email: str,
    subject: str,
    template: Union[str, LazyI18nString],
    context: Dict[str, Any] = None,
    locale: str = None,
    headers: dict = None,
):
    headers = headers or {}
    c = Configuration.get_solo()
    locale = locale or c.language

    with override(locale):
        body = str(template)
        if context:
            body = body.format_map(TolerantDict(context))

        sender = Configuration.get_solo().mail_from
        subject = str(subject)
        body_plain = body
        return mail_send_task.apply_async(
            args=([email], subject, body_plain, sender, headers)
        )


def mail_send_task(
    to: str,
    subject: str,
    body: str,
    sender: str,
    html: str = None,
    cc: list = None,
    bcc: list = None,
    headers: dict = None,
    attachments: list = None,
):
    email = EmailMultiAlternatives(
        subject, body, sender, to=to, cc=cc, bcc=bcc, headers=headers
    )
    if html is not None:
        email.attach_alternative(inline_css(html), "text/html")
    if attachments:
        from byro.documents.models import Document

        for attachment in attachments:
            email.attach_file(Document.objects.get(pk=attachment).document.path)
    backend = get_connection(fail_silently=False)

    try:
        backend.send_messages([email])
    except Exception:
        logger.exception("Error sending email")
        raise SendMailException(f"Failed to send an email to {to}.")



# Generated by Django 1.11.4 on 2017-08-12 12:37

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("members", "0001_initial"),
        ("bookkeeping", "0002_auto_20170812_1109"),
    ]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="member",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="accounts",
                to="members.Member",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="account",
            unique_together={("member", "account_category", "name")},
        ),
    ]



# Generated by Django 1.11.8 on 2018-02-24 20:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("mails", "0002_email_attachments")]

    operations = [
        migrations.AddField(
            model_name="mailtemplate",
            name="reply_to",
            field=models.EmailField(
                blank=True,
                help_text="Change the Reply-To address if you do not want to use the default orga address",
                max_length=200,
                null=True,
                verbose_name="Reply-To",
            ),
        )
    ]



from django import forms
from django.contrib import messages
from django.db import models
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, FormView, ListView

from byro.bookkeeping.models import Account, AccountCategory, Transaction

FORM_CLASS = forms.modelform_factory(Account, fields=["name", "account_category"])

ACCOUNT_COLUMN_HEADERS = {
    # FIXME Check this with an accountant who is a native english speaker
    AccountCategory.INCOME: (_("Charge"), _("Revenue")),
    AccountCategory.ASSET: (_("Increase"), _("Decrease")),
    AccountCategory.EQUITY: (_("Decrease"), _("Increase")),
    AccountCategory.LIABILITY: (_("Decrease"), _("Increase")),
    AccountCategory.EXPENSE: (_("Expense"), _("Rebate")),
}


class AccountListView(ListView):
    template_name = "office/account/list.html"
    context_object_name = "accounts"
    model = Account


class AccountCreateView(FormView):
    template_name = "office/account/add.html"
    model = Account
    form_class = FORM_CLASS

    def form_valid(self, form):
        form.save()
        messages.success(
            self.request,
            _("The account was added, please edit additional details if applicable."),
        )
        self.form = form
        return super().form_valid(form)

    def get_success_url(self):
        return reverse(
            "office:finance.accounts.detail", kwargs={"pk": self.form.instance.pk}
        )


class AccountDetailView(ListView):
    template_name = "office/account/detail.html"
    context_object_name = "bookings"
    model = Transaction
    paginate_by = 25

    def get_object(self):
        if not hasattr(self, "object"):
            self.object = Account.objects.get(pk=self.kwargs["pk"])
        return self.object

    def get_queryset(self):
        qs = self.get_object().bookings_with_transaction_data
        if self.request.GET.get("filter") == "unbalanced":
            qs = qs.exclude(
                transaction_balances_debit=models.F("transaction_balances_credit")
            )
        qs = qs.filter(transaction__value_datetime__lte=now()).order_by(
            "-transaction__value_datetime"
        )
        return qs

    def get_form(self, request=None):
        form = FORM_CLASS(request.POST if request else None, instance=self.get_object())
        form.fields["account_category"].disabled = True
        return form

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["form"] = self.get_form()
        context["account"] = self.get_object()
        context["ACCOUNT_COLUMN_HEADERS"] = ACCOUNT_COLUMN_HEADERS.get(
            self.get_object().account_category, (_("Debit"), _("Credit"))
        )
        return context

    def post(self, request, *args, **kwargs):
        form = self.get_form(request)
        if form.is_valid() and form.has_changed():
            form.save()
            messages.success(self.request, _("Your changes have been saved."))
        return redirect(reverse("office:finance.accounts.detail", kwargs=self.kwargs))


class AccountDeleteView(DetailView):
    model = Account
    context_object_name = "account"



# Generated by Django 1.11.8 on 2018-02-24 21:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("mails", "0003_mailtemplate_reply_to"),
        ("common", "0005_configuration_liability_interval"),
    ]

    operations = [
        migrations.AddField(
            model_name="configuration",
            name="leave_member_template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="mails.MailTemplate",
            ),
        ),
        migrations.AddField(
            model_name="configuration",
            name="leave_office_template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="mails.MailTemplate",
            ),
        ),
        migrations.AddField(
            model_name="configuration",
            name="welcome_member_template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="mails.MailTemplate",
            ),
        ),
        migrations.AddField(
            model_name="configuration",
            name="welcome_office_template",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="mails.MailTemplate",
            ),
        ),
    ]



from decimal import Decimal

from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils.functional import classproperty
from django.utils.safestring import mark_safe
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from byro.common.models import LogTargetMixin
from byro.common.models.auditable import Auditable
from byro.common.models.choices import Choices


class AccountCategory(Choices):
    # Categories for double-entry bookkeeping
    ASSET = "asset"  # de: Aktiva, for example your bank account or cash
    LIABILITY = "liability"  # de: Passiva, for example invoices you have to pay
    INCOME = "income"  # de: Ertragskonten, for example for fees paid
    EXPENSE = "expense"  # de: Aufwandskonten, for example for fees to be paid
    EQUITY = "equity"  # de: Eigenkapital, assets less liabilities

    @classproperty
    def choices(cls):
        return (
            (cls.ASSET, _("Asset account")),
            (cls.LIABILITY, _("Liability account")),
            (cls.INCOME, _("Income account")),
            (cls.EXPENSE, _("Expense account")),
            (cls.EQUITY, _("Equity account")),
        )


class AccountTag(models.Model):
    name = models.CharField(max_length=300, null=False, unique=True)
    description = models.CharField(max_length=1000, null=True)

    def __str__(self):
        return self.name


class Account(Auditable, models.Model, LogTargetMixin):
    account_category = models.CharField(
        choices=AccountCategory.choices, max_length=AccountCategory.max_length
    )
    name = models.CharField(max_length=300, null=True)  # e.g. 'Laser donations'
    tags = models.ManyToManyField(AccountTag, related_name="accounts")

    class Meta:
        unique_together = (("account_category", "name"),)

    def __str__(self):
        if self.name:
            return self.name
        return "{self.account_category} account #{self.id}".format(self=self)

    @property
    def bookings(self):
        from byro.bookkeeping.models import Booking

        return Booking.objects.filter(Q(debit_account=self) | Q(credit_account=self))

    @property
    def bookings_with_transaction_data(self):
        from byro.bookkeeping.models import Booking

        return Booking.objects.with_transaction_data().filter(
            Q(debit_account=self) | Q(credit_account=self)
        )

    @property
    def transactions(self):
        from byro.bookkeeping.models import Transaction

        return Transaction.objects.filter(
            Q(bookings__debit_account=self) | Q(bookings__credit_account=self)
        )

    @property
    def unbalanced_transactions(self):
        from byro.bookkeeping.models import Transaction

        return Transaction.objects.unbalanced_transactions().filter(
            Q(bookings__debit_account=self) | Q(bookings__credit_account=self)
        )

    def _filter_by_date(self, qs, start, end):
        if start:
            qs = qs.filter(value_datetime__gte=start)
        if end:
            qs = qs.filter(value_datetime__lte=end)
        return qs

    def balances(self, start=None, end=None):
        end = end or now()
        qs = self._filter_by_date(self.transactions, start, end)

        result = qs.with_balances().aggregate(
            debit=models.functions.Coalesce(
                models.Sum("balances_debit"), 0, output_field=models.DecimalField()
            ),
            credit=models.functions.Coalesce(
                models.Sum("balances_credit"), 0, output_field=models.DecimalField()
            ),
        )

        # ASSET, EXPENSE:  Debit increases balance, credit decreases it
        # INCOME, LIABILITY, EQUITY:  Debit decreases balance, credit increases it

        if self.account_category in (
            AccountCategory.LIABILITY,
            AccountCategory.INCOME,
            AccountCategory.EQUITY,
        ):
            result["net"] = result["credit"] - result["debit"]
        else:
            result["net"] = result["debit"] - result["credit"]

        result = {k: Decimal(v).quantize(Decimal("0.01")) for k, v in result.items()}

        return result

    def get_absolute_url(self):
        return reverse("office:finance.accounts.detail", kwargs={"pk": self.pk})

    def get_object_icon(self):
        return mark_safe('<i class="fa fa-bank"></i> ')



from datetime import timedelta

from django import forms
from django.db.models.fields.related import OneToOneRel
from django.utils.timezone import now

from byro.common.forms.registration import DefaultDates
from byro.common.models import Configuration
from byro.members.models import Member, Membership, get_next_member_number

MAPPING = {"member": Member, "membership": Membership}


class CreateMemberForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        config = Configuration.get_solo().registration_form or []
        config = sorted(
            (field for field in config if field["position"] is not None),
            key=lambda field: field["position"],
        )
        profiles = {
            profile.related_model.__name__: profile.related_model
            for profile in Member._meta.related_objects
            if isinstance(profile, OneToOneRel) and profile.name.startswith("profile_")
        }
        for field in config:
            self.build_field(field, profiles)
            if field.get("name", "") == "member__name":
                self.build_field({"name": "member__direct_address_name"}, [])
                self.build_field({"name": "member__order_name"}, [])

        if "member__number" in self.fields:
            self.fields["member__number"].initial = get_next_member_number()

    def get_date_initial(self, field):
        today = now().date()
        default = field["default_date"]
        if default == DefaultDates.TODAY:
            return today
        elif default == DefaultDates.BEGINNING_MONTH:
            return today.replace(day=1)
        elif default == DefaultDates.BEGINNING_MONTH_NEXT:
            return (today.replace(day=28) + timedelta(days=7)).replace(day=1)
        elif default == DefaultDates.BEGINNING_YEAR:
            return today.replace(day=1, month=1)
        elif default == DefaultDates.BEGINNING_YEAR_NEXT:
            return (today.replace(day=31, month=12) + timedelta(days=7)).replace(
                day=1, month=1
            )
        elif default == DefaultDates.FIXED_DATE:
            return field.get("default", None)

    def build_field(self, field, profiles):
        model_name = field["name"].split("__")[0]
        if model_name in MAPPING:
            model = MAPPING[model_name]
        else:
            model = profiles[field["name"].split("__")[0]]
        form_field = model._meta.get_field(field["name"].split("__")[-1]).formfield()
        form_field.model = model
        self.fields[field["name"]] = form_field
        if "default_date" in field:
            form_field.initial = self.get_date_initial(field=field)
        elif "default_boolean" in field:
            form_field.initial = field["default_boolean"]
        elif "default" in field:
            form_field.initial = field["default"]

    def save(self):
        profiles = {
            profile.related_model.__name__: profile.related_model()
            for profile in Member._meta.related_objects
            if isinstance(profile, OneToOneRel) and profile.name.startswith("profile_")
        }
        member = Member()
        membership = Membership()

        for key, value in self.cleaned_data.items():
            model_name = key.split("__")[0]
            if model_name == "member":
                obj = member
            elif model_name == "membership":
                obj = membership
            else:
                obj = profiles[key.split("__")[0]]
            setattr(obj, key.split("__", maxsplit=1)[-1], value)
        member.save()
        member.refresh_from_db()
        membership.member = self.instance = member
        membership.save()
        for profile in profiles.values():
            profile.member = member
            profile.save()



import base64
import io

import qrcode
from django import template

register = template.Library()


@register.filter
def qrcode_inline(value):
    qr = qrcode.QRCode(
        version=None, box_size=2, error_correction=qrcode.constants.ERROR_CORRECT_Q
    )
    qr.add_data(value)
    qr.make(fit=True)
    img = qr.make_image()
    with io.BytesIO() as output:
        img.save(output, format="PNG")
        return "data:image/png;base64,{}".format(
            base64.b64encode(output.getvalue()).decode("us-ascii")
        )



# Generated by Django 1.11.10 on 2018-02-25 20:05

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("common", "0007_auto_20180224_2114")]

    operations = [
        migrations.AddField(
            model_name="configuration",
            name="backoffice_mail",
            field=models.EmailField(
                blank=True,
                max_length=100,
                null=True,
                verbose_name="e-mail of backoffice",
            ),
        )
    ]



# Generated by Django 1.11.8 on 2018-01-13 14:02

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("common", "0004_auto_20180111_1807")]

    operations = [
        migrations.AddField(
            model_name="configuration",
            name="liability_interval",
            field=models.IntegerField(
                default=36,
                help_text="For which interval should remaining fees be calculated?",
                verbose_name="Liability interval",
            ),
        )
    ]



import re
import sys

import canonicaljson
from django.core.management.base import BaseCommand

from byro.common.models.log import LogEntry


class Command(BaseCommand):
    help = "Export the log-chain in JSON format"

    def add_arguments(self, parser):
        parser.add_argument(
            "-a",
            "--data-include-actions",
            default=r"^.*$",
            type=str,
            help="action_types for which to include the log 'data' member (Regex)",
        )
        parser.add_argument(
            "-A",
            "--data-exclude-actions",
            default=None,
            type=str,
            help="action_types for which to exclude the log 'data' member (Regex)",
        )

    def handle(self, *args, **options):
        include_re = re.compile(options["data_include_actions"])
        exclude_re = (
            re.compile(options["data_exclude_actions"])
            if options["data_exclude_actions"]
            else None
        )

        outstream = sys.stdout

        outstream.write("[")
        is_first = True
        last = None
        current = LogEntry.objects.get_chain_end()

        while current and current != last:
            data = {
                "hash": current.auth_hash,
                "entry": current.get_authenticated_dict(),
            }

            if include_re.search(current.action_type) and (
                exclude_re is None or not exclude_re.search(current.action_type)
            ):
                data["data"] = current.data

            data_pretty = canonicaljson.encode_pretty_printed_json(data).decode("utf-8")

            if not is_first:
                outstream.write(",\n    ")
            else:
                outstream.write("\n    ")

            outstream.write("\n    ".join(data_pretty.split("\n")))

            is_first = False
            last = current
            current = current.auth_prev

        outstream.write("\n]\n")



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DocumentsConfig(AppConfig):
    name = "byro.documents"

    class ByroPluginMeta:
        document_categories = {
            "byro.documents.misc": _("Miscellaneous document"),
            "byro.documents.registration_form": _("Registration form"),
        }



from django.core.management.base import BaseCommand

from byro.common.signals import periodic_task


class Command(BaseCommand):
    help = "Run periodic tasks"

    def handle(self, *args, **options):
        periodic_task.send(self)



from contextlib import suppress
from datetime import datetime

from django.utils.timezone import now
from more_itertools import peekable

from byro.bookkeeping.models import Booking
from byro.bookkeeping.special_accounts import SpecialAccounts


def _date_compare_key(d):
    # Returns datetime as is, Converts date to datetime at end of day local timezone
    if isinstance(d, datetime):
        return d
    else:
        return now().replace(d.year, d.month, d.day, 23, 59, 59, 999999)


def sorted_merge(*args):
    initer = [iter(e) for e in args]
    nextin = [None] * len(initer)

    for i, e in enumerate(initer):
        with suppress(StopIteration):
            nextin[i] = next(e)

    while any(e is not None for e in nextin):
        available = [i for (i, e) in enumerate(nextin) if e is not None]
        if len(available) > 1:
            minindex = max(
                *available, key=lambda a: _date_compare_key(nextin[a]["date"])
            )
        else:
            minindex = available[0]
        yield nextin[minindex]
        nextin[minindex] = None
        with suppress(StopIteration):
            nextin[minindex] = next(initer[minindex])


def get_mail_timeline(member):
    for instance in member.emails.filter(sent__isnull=False).order_by("-sent").all():
        yield {
            "type": "mail",
            "subtype": "mail-out",
            "date": instance.sent,
            "icon": "envelope",
            "instance": instance,
        }


def get_base_finance_timeline(member):
    last_transaction_pk = None
    for instance in (
        Booking.objects.with_transaction_data()
        .filter(member=member, transaction__reverses__isnull=True)
        .order_by("-transaction__value_datetime", "-transaction__pk")
    ):
        if instance.transaction.pk == last_transaction_pk:
            continue
        last_transaction_pk = instance.transaction.pk
        base_data = {
            "type": "finance",
            "date": instance.transaction.value_datetime,
            "icon": "money",
            "instance": instance,
            "deleted": instance.transaction.reversed_by.first(),
        }
        b = list(instance.transaction.bookings.all())
        if any(
            e.member == member and e.debit_account == SpecialAccounts.fees_receivable
            for e in b
        ) and any(
            e.member == member and e.credit_account == SpecialAccounts.fees for e in b
        ):
            yield dict(subtype="membership-due", value=instance.amount, **base_data)
        elif any(
            e.member == member and e.credit_account == SpecialAccounts.fees_receivable
            for e in b
        ):
            yield dict(subtype="membership-paid", value=instance.amount, **base_data)
        else:
            yield dict(subtype="other-transaction", value=instance.amount, **base_data)


def get_misc_finance_timeline(member):
    for entry in member.log_entries().filter(
        action_type__startswith="byro.members.finance."
    ):
        base_data = {
            "type": "finance",
            "icon": "money",
            "instance": entry,
            "date": entry.datetime,
        }
        if (
            entry.action_type
            == "byro.members.finance.sepadd.mandate_reference_assigned"
        ):
            yield dict(
                base_data, subtype="sepadd-mandate-reference-assigned", icon="info"
            )
        else:
            yield dict(base_data, subtype="other")


def get_finance_timeline(member):
    return sorted_merge(
        get_base_finance_timeline(member), get_misc_finance_timeline(member)
    )


def get_misc_ops_timeline(member):
    for entry in member.log_entries().exclude(
        action_type__startswith="byro.members.finance."
    ):
        base_data = {
            "type": "ops",
            "icon": "user",
            "instance": entry,
            "date": entry.datetime,
        }
        if entry.action_type == "byro.members.created":
            yield dict(base_data, subtype="member-created", icon="user-plus")
        elif entry.action_type == "byro.members.updated":
            yield dict(base_data, subtype="member-updated", icon="pencil")
        elif entry.action_type == "byro.members.document.created":
            yield dict(base_data, subtype="document-created", icon="file")
        else:
            yield dict(base_data, subtype="other")


def get_ops_timeline(member):
    membership_ops = []
    for membership in member.memberships.all():
        base_data = {"type": "ops", "icon": "user", "instance": membership}
        if membership.start:
            membership_ops.append(
                dict(subtype="membership-begin", date=membership.start, **base_data)
            )
        if membership.end:
            membership_ops.append(
                dict(subtype="membership-end", date=membership.end, **base_data)
            )

    return sorted_merge(
        sorted(membership_ops, reverse=True, key=lambda a: a["date"]),
        get_misc_ops_timeline(member),
    )


def get_file_icon(document):
    return {"application/pdf": "file-pdf-o"}.get(document.mime_type_guessed, "file-o")


def get_document_timeline(member):
    for document in member.documents.order_by("-date", "-id").all():
        base_data = {
            "type": "document",
            "icon": get_file_icon(document),
            "instance": document,
            "date": document.date,
        }
        if document.category == "byro.documents.registration_form":
            yield dict(base_data, subtype="registration_form")
        else:
            yield dict(base_data, subtype="misc_document")


def add_dummy_entries(entries):
    entries = peekable(entries)
    prev_entry = next(entries)
    yield prev_entry
    output_month, output_year = prev_entry["date"].month, prev_entry["date"].year

    while entries:
        entry = next(entries)
        while output_month != entry["date"].month or output_year != entry["date"].year:
            output_month = output_month - 1
            if output_month < 1:
                output_month = 12
                output_year = output_year - 1
            if output_month != entry["date"].month or output_year != entry["date"].year:
                yield {
                    "type": "dummy",
                    "subtype": "dummy",
                    "instance": None,
                    "date": datetime(year=output_year, month=output_month, day=1),
                }
        yield entry


def augment_timeline(entries):
    last_year = None
    last_month = None

    entries = peekable(add_dummy_entries(entries))

    while entries:
        entry = next(entries)
        nentry = entries.peek(None)
        tl = {
            "year_first": False,
            "year_last": False,
            "month_first": False,
            "month_last": False,
            "entry_id": "{}:{}:{}".format(
                entry["type"], entry["subtype"], entry["instance"].pk
            )
            if entry["instance"]
            else None,
        }
        if last_year != entry["date"].year:
            tl["year_first"] = True
        if last_month != entry["date"].month:
            tl["month_first"] = True
        if not nentry or nentry["date"].year != entry["date"].year:
            tl["year_last"] = True
        if not nentry or nentry["date"].month != entry["date"].month:
            tl["month_last"] = True
        yield dict(entry, tl=tl)
        last_year = entry["date"].year
        last_month = entry["date"].month



# Generated by Django 2.1.1 on 2018-09-11 19:56

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("members", "0009_auto_20180512_1810"),
        ("mails", "0004_email_template"),
    ]

    operations = [
        migrations.AddField(
            model_name="email",
            name="members",
            field=models.ManyToManyField(
                null=True, related_name="emails", to="members.Member"
            ),
        )
    ]



from collections import OrderedDict
from decimal import Decimal
from functools import reduce

from dateutil.relativedelta import relativedelta
from django.db import models, transaction
from django.db.models import Q
from django.db.models.fields.related import OneToOneRel
from django.urls import reverse
from django.utils.functional import classproperty
from django.utils.safestring import mark_safe
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from byro.bookkeeping.models import Booking, Transaction
from byro.bookkeeping.special_accounts import SpecialAccounts
from byro.common.models import Configuration, LogEntry, LogTargetMixin
from byro.common.models.auditable import Auditable
from byro.common.models.choices import Choices
from byro.common.templatetags.format_with_currency import format_with_currency


class Field:
    field_id = str
    name = str
    description = str

    computed = bool
    read_only = bool
    registration_form = dict

    def __init__(
        self,
        field_id,
        name,
        description,
        path,
        addition=None,
        registration_form=None,
        **kwargs,
    ):
        self.field_id = field_id
        self.base_name = name
        if addition:
            self.name = f"{name} ({addition})"
        else:
            self.name = name
        self.description = description
        self.path = path
        self.registration_form = registration_form or {}
        for k, v in kwargs.items():
            setattr(self, k, v)

    @staticmethod
    def _follow_path(start, path):
        """Follow a python.dot.path until its penultimate item, then return the
        current object and the name of the last descriptor. Allows 'func()'
        syntax to call a method along the way without arguments.

        Examples:
            _follow_path(m, 'pk')  ->  m, 'pk'
            _follow_path(m, 'profile_foo.bar')  ->  m.profile_foo, 'bar'
            _follow_path(m, 'memberships.last().start')  ->  m.memberships.last(), 'start'
        """
        target = start
        path = path.split(".")
        for p in path[:-1]:
            target = getattr(target, p.rsplit("(")[0], None)
            if p.endswith("()") and target is not None:
                target = target()
        return target, path[-1]

    def getter(self, member):
        target, prop = self._follow_path(member, self.path)
        return getattr(target, prop, None)

    def setter(self, member, value):
        if self.read_only:
            raise NotImplementedError(f"Writing to {self.path} is not supported")
        target, prop = self._follow_path(member, self.path)
        if target is None:
            raise AttributeError(f"Encountered 'None' while following {self.path}")
        setattr(target, prop, value)
        if callable(getattr(target, "save", None)):
            target.save()


class MemberTypes:
    MEMBER = "member"
    EXTERNAL = "external"


class MemberContactTypes(Choices):
    ORGANIZATION = "organization"
    PERSON = "person"
    ROLE = "role"


class MemberQuerySet(models.QuerySet):
    def with_active_membership(self):
        return (
            self.filter(
                Q(memberships__start__lte=now().date())
                & (
                    Q(memberships__end__isnull=True)
                    | Q(memberships__end__gte=now().date())
                )
            )
            .order_by("-id")
            .distinct()
        )


class MemberManager(models.Manager):
    def get_queryset(self):
        return MemberQuerySet(self.model, using=self._db).filter(
            membership_type=MemberTypes.MEMBER
        )

    def with_active_membership(self):
        return self.get_queryset().with_active_membership()


class AllMemberManager(models.Manager):
    pass


def get_next_member_number():
    all_numbers = Member.all_objects.all().values_list("number", flat=True)
    numeric_numbers = [n for n in all_numbers if n is not None and n.isdigit()]
    try:
        return max(int(n) for n in numeric_numbers) + 1
    except Exception:
        return 1


def get_member_data(obj):
    if hasattr(obj, "get_member_data"):
        return obj.get_member_data()
    return [
        (field.verbose_name, str(getattr(obj, field.name)))
        for field in obj._meta.fields
        if field.name
        not in ("id", "created_by", "modified_by", "created", "modified", "member")
    ]


class Member(Auditable, models.Model, LogTargetMixin):
    LOG_TARGET_BASE = "byro.members"

    number = models.CharField(
        max_length=100,
        verbose_name=_("Membership number/ID"),
        null=True,
        blank=True,
        db_index=True,
    )
    name = models.CharField(
        max_length=100, verbose_name=_("Name"), null=True, blank=True
    )
    direct_address_name = models.CharField(
        max_length=100, verbose_name=_("Name for direct address"), null=True, blank=True
    )
    order_name = models.CharField(
        max_length=100, verbose_name=_("Name (sort order)"), null=True, blank=True
    )
    address = models.TextField(
        max_length=300, verbose_name=_("Address"), null=True, blank=True
    )
    email = models.EmailField(
        max_length=200, verbose_name=_("E-Mail"), null=True, blank=True
    )
    membership_type = models.CharField(max_length=40, default=MemberTypes.MEMBER)
    member_contact_type = models.CharField(
        max_length=MemberContactTypes.max_length,
        verbose_name=_("Contact type"),
        choices=MemberContactTypes.choices,
        default=MemberContactTypes.PERSON,
    )

    form_title = _("Member")
    objects = MemberManager()
    all_objects = AllMemberManager()

    class Meta:
        ordering = ("direct_address_name", "name")

    @classproperty
    def profile_classes(cls) -> list:
        return [
            related.related_model
            for related in cls._meta.related_objects
            if isinstance(related, OneToOneRel) and related.name.startswith("profile_")
        ]

    @property
    def profiles(self) -> list:
        return [
            getattr(self, related.name)
            for related in self._meta.related_objects
            if isinstance(related, OneToOneRel) and related.name.startswith("profile_")
        ]

    @classmethod
    def get_query_for_search(cls, search):
        search_parts = search.strip().split()
        query = models.Q()
        for part in search_parts:
            query |= models.Q(name__icontains=part)
            query |= models.Q(profile_profile__nick__icontains=part)
            query |= models.Q(number=search)
        return query

    @classmethod
    def get_fields(cls):
        result = []

        result.append(
            Field(
                "_internal_id",
                _("Internal database ID"),
                "",
                "pk",
                computed=True,
                read_only=True,
            )
        )
        result.append(
            Field(
                "_internal_active",
                _("Member active?"),
                "",
                "is_active",
                computed=True,
                read_only=True,
            )
        )
        result.append(
            Field(
                "_internal_balance",
                _("Account balance"),
                "",
                "balance",
                computed=True,
                read_only=True,
            )
        )
        result.append(
            Field(
                "_internal_last_transaction",
                _("Last Member Fee Transaction Timestamp"),
                "",
                "last_membership_fee_transaction_timestamp",
                computed=True,
                read_only=True,
            )
        )

        reg_form = Configuration.get_solo().registration_form or []
        form_config = {entry["name"]: entry for entry in reg_form}

        profile_map = {
            profile.related_model: profile.name
            for profile in cls._meta.related_objects
            if isinstance(profile, OneToOneRel) and profile.name.startswith("profile_")
        }

        for model in [cls, Membership] + cls.profile_classes:
            for field in model._meta.fields:
                if field.name in ("id", "member") or (
                    model is Member and field.name == "membership_type"
                ):
                    continue

                f_id = "{}__{}".format(
                    SPECIAL_NAMES.get(model, model.__name__), field.name
                )
                f_name = field.verbose_name or field.name
                f_addition = None

                if issubclass(model, cls):
                    f_path = field.name
                elif model is Membership:
                    f_path = f"memberships.last().{field.name}"
                    f_addition = _("Current membership")
                else:
                    f_path = f"{profile_map[model]}.{field.name}"
                    f_addition = model.__name__

                result.append(
                    Field(
                        f_id,
                        f_name,
                        "",
                        f_path,
                        addition=f_addition,
                        registration_form=form_config.get(f_id, None),
                        computed=False,
                        read_only=False,
                    )
                )

        return OrderedDict([(f.field_id, f) for f in result])

    def _calc_balance(
        self,
        liability_cutoff=None,
        asset_cutoff=None,
        liability_start=None,
        asset_start=None,
    ) -> Decimal:
        _now = now()
        fees_receivable_account = SpecialAccounts.fees_receivable
        qs = Booking.objects.filter(member=self)
        debits = qs.filter(
            debit_account=fees_receivable_account,
            transaction__value_datetime__lte=liability_cutoff or _now,
        )
        if liability_start:
            debits = debits.filter(transaction__value_datetime__gte=liability_start)
        credits = qs.filter(
            credit_account=fees_receivable_account,
            transaction__value_datetime__lte=asset_cutoff or _now,
        )
        if asset_start:
            credits = credits.filter(transaction__value_datetime__gte=asset_start)
        liability = debits.aggregate(liability=models.Sum("amount"))[
            "liability"
        ] or Decimal("0.00")
        asset = credits.aggregate(asset=models.Sum("amount"))["asset"] or Decimal(
            "0.00"
        )
        return asset - liability

    @property
    def balance(self) -> Decimal:
        return self._calc_balance()

    def _calc_last_membership_fee_transaction_timestamp(self):
        _now = now()
        fees_receivable_account = SpecialAccounts.fees_receivable
        qs = Booking.objects.filter(
            Q(debit_account=fees_receivable_account)
            | Q(credit_account=fees_receivable_account),
            member=self,
            transaction__value_datetime__lte=_now,
        )
        return qs.aggregate(last_transaction=models.Max("transaction__value_datetime"))[
            "last_transaction"
        ]

    @property
    def last_membership_fee_transaction_timestamp(self):
        return self._calc_last_membership_fee_transaction_timestamp()

    def create_balance(self, start, end, commit=True, create_if_zero=True):
        if self.balances.exists():
            if self.balances.filter(
                models.Q(
                    models.Q(start__lte=start) & models.Q(end__gte=start)
                )  # start overlaps
                | models.Q(
                    models.Q(start__lte=end) & models.Q(end__gte=end)
                )  # end overlaps
            ).exists():
                raise Exception(
                    "Cannot create overlapping balance: {} from {} to {}".format(
                        self, start, end
                    )
                )
        amount = self._calc_balance(
            liability_cutoff=end,
            asset_cutoff=end,
            liability_start=start,
            asset_start=start,
        )
        if not amount and not create_if_zero:
            return
        balance = MemberBalance(member=self, start=start, end=end, amount=amount)
        if commit is True:
            balance.save()
        return balance

    def statute_barred_debt(self, future_limit=None) -> Decimal:
        future_limit = future_limit or relativedelta()
        limit = (
            relativedelta(months=Configuration.get_solo().liability_interval)
            - future_limit
        )
        last_unenforceable_date = (
            now().replace(month=12, day=31) - limit - relativedelta(years=1)
        )
        return max(Decimal("0.00"), -self._calc_balance(last_unenforceable_date))

    @property
    def donation_balance(self) -> Decimal:
        return self.donations.aggregate(donations=models.Sum("amount"))[
            "donations"
        ] or Decimal("0.00")

    @property
    def donations(self):
        return Booking.objects.filter(
            credit_account=SpecialAccounts.donations,
            member=self,
            transaction__value_datetime__lte=now(),
        )

    @property
    def fee_payments(self):
        return Booking.objects.filter(
            debit_account=SpecialAccounts.fees_receivable,
            member=self,
            transaction__value_datetime__lte=now(),
        )

    @property
    def is_active(self):
        if not self.memberships.count():
            return False
        result = self.memberships.last().start <= now().date()
        if self.memberships.last().end:
            result = result and not (self.memberships.last().end < now().date())
        return result

    @property
    def record_disclosure_email(self):
        config = Configuration.get_solo()
        template = config.record_disclosure_template
        data = get_member_data(self)
        for profile in self.profiles:
            data += get_member_data(profile)
        key_value_data = [d for d in data if len(d) == 2 and not isinstance(d, str)]
        text_data = [d for d in data if isinstance(d, str)]
        key_length = min(max(len(d[0]) for d in key_value_data), 20)
        key_value_text = []
        for key, value in key_value_data:
            key = (key + ":").ljust(key_length) + " "
            value = value.strip()
            if value in [None, "None", ""]:
                value = "-"
            if isinstance(value, str) and "\n" in value:
                value = "\n".join(
                    [" " * (key_length + 1) + line for line in value.split("\n")]
                ).strip()
            key_value_text.append(key + value)
        key_value_text = "\n".join(key_value_text)
        if text_data:
            key_value_text += "\n" + "\n".join(text_data)
        context = {
            "association_name": config.name,
            "data": key_value_text,
            "number": self.number,
            "balance": format_with_currency(self.balance, long_form=True),
        }
        return template.to_mail(self.email, context=context, save=False)

    @transaction.atomic
    def update_liabilites(self):
        src_account = SpecialAccounts.fees
        dst_account = SpecialAccounts.fees_receivable

        config = Configuration.get_solo()

        # Step 1: Identify all dates and amounts that should be due at those dates
        #  (in python, store as a list; hits database once to get list of memberships)
        # Step 2: Find all due amounts within the data ranges, ignore reversed liabilities
        #  (hits database)
        # Step 3: Compare due date and amounts with list from step 1
        #  (in Python)
        # Step 4: Cancel all liabilities that didn't match in step 3
        #  (hits database, once per mismatch)
        # Step 5: Add all missing liabilities
        #  (hits database, once per new due)
        # Step 6: Find and cancel all liabilities outside of membership dates, replaces remove_future_liabilites_on_leave()
        #  (hits database, once for search, once per stray liability)

        dues = set()
        membership_ranges = []
        _now = now()
        _from = config.accounting_start

        # Step 1
        for membership in self.memberships.all():
            if not membership.amount:
                continue
            membership_range, membership_dues = membership.get_dues(
                _now=_now, _from=_from
            )
            membership_ranges.append(membership_range)
            dues |= membership_dues

        # Step 2
        dues_qs = Booking.objects.filter(
            member=self,
            credit_account=src_account,
            transaction__reversed_by__isnull=True,
        )
        if membership_ranges:
            date_range_q = reduce(
                lambda a, b: a | b,
                [
                    models.Q(transaction__value_datetime__gte=start)
                    & models.Q(transaction__value_datetime__lte=end)
                    for start, end in membership_ranges
                ],
            )
            dues_qs = dues_qs.filter(date_range_q)
        dues_in_db = {  # Must be a dictionary instead of set, to retrieve b later on
            (b.transaction.value_datetime.date(), b.amount): b for b in dues_qs.all()
        }

        # Step 3
        dues_in_db_as_set = set(dues_in_db.keys())
        wrong_dues_in_db = dues_in_db_as_set - dues
        missing_dues = dues - dues_in_db_as_set

        # Step 4
        for wrong_due in sorted(wrong_dues_in_db):
            booking = dues_in_db[wrong_due]
            booking.transaction.reverse(
                memo=_("Due amount canceled because of change in membership amount"),
                user_or_context="internal: update_liabilites, membership amount changed",
            )

        # Step 5:
        for date, amount in sorted(missing_dues):
            t = Transaction.objects.create(
                value_datetime=date,
                booking_datetime=_now,
                memo=_("Membership due"),
                user_or_context="internal: update_liabilites, add missing liabilities",
            )
            t.credit(
                account=src_account,
                amount=amount,
                member=self,
                user_or_context="internal: update_liabilites, add missing liabilities",
            )
            t.debit(
                account=dst_account,
                amount=amount,
                member=self,
                user_or_context="internal: update_liabilites, add missing liabilities",
            )
            t.save()

        # Step 6:
        stray_liabilities_qs = Booking.objects.filter(
            member=self,
            credit_account=src_account,
            transaction__reversed_by__isnull=True,
        )
        if membership_ranges:
            stray_liabilities_qs = stray_liabilities_qs.exclude(date_range_q)
        stray_liabilities_qs = stray_liabilities_qs.prefetch_related("transaction")
        for stray_liability in stray_liabilities_qs.all():
            stray_liability.transaction.reverse(
                memo=_("Due amount outside of membership canceled"),
                user_or_context="internal: update_liabilites, reverse stray liabilities",
            )

    @transaction.atomic
    def adjust_balance(self, user_or_context, memo, amount, from_, to_, value_datetime):
        now_ = now()

        if not amount:
            return False

        if amount < 0:
            amount = -amount
            from_, to_ = to_, from_

        from_member = self if from_ == SpecialAccounts.fees_receivable else None
        to_member = self if to_ == SpecialAccounts.fees_receivable else None

        t = Transaction.objects.create(
            value_datetime=value_datetime,
            booking_datetime=now_,
            user_or_context=user_or_context,
            memo=memo,
        )
        t.debit(
            account=to_,
            member=to_member,
            amount=amount,
            user_or_context=user_or_context,
        )
        t.credit(
            account=from_,
            member=from_member,
            amount=amount,
            user_or_context=user_or_context,
        )

        return True

    def update_name_fields(self, force=False, save=True):
        self.update_order_name(force=force, save=save)
        self.update_direct_address_name(force=force, save=save)

    def update_order_name(self, force=False, save=True):
        if not self.order_name or force:
            name_parts = self.name.split()
            config = Configuration.get_solo()
            self.order_name = (
                name_parts[0]
                if config.default_order_name == "first"
                else name_parts[-1]
            )
            if save:
                self.save()

    def update_direct_address_name(self, force=False, save=True):
        if not self.direct_address_name or force:
            name_parts = self.name.split()
            config = Configuration.get_solo()
            self.direct_address_name = (
                name_parts[0]
                if config.default_direct_address_name == "first"
                else name_parts[-1]
            )
            if save:
                self.save()

    def __str__(self):
        return "Member {self.number} ({self.name})".format(self=self)

    def get_absolute_url(self):
        return reverse("office:members.data", kwargs={"pk": self.pk})

    def get_object_icon(self):
        return mark_safe('<i class="fa fa-user"></i> ')

    def log_entries(self):
        own_entries = [e.pk for e in super().log_entries()]
        ms_entries = [e.pk for m in self.memberships.all() for e in m.log_entries()]
        return LogEntry.objects.filter(pk__in=own_entries + ms_entries)


class FeeIntervals:
    MONTHLY = 1
    QUARTERLY = 3
    BIANNUAL = 6
    ANNUALLY = 12

    @classproperty
    def choices(cls):
        return (
            (cls.MONTHLY, _("monthly")),
            (cls.QUARTERLY, _("quarterly")),
            (cls.BIANNUAL, _("biannually")),
            (cls.ANNUALLY, _("annually")),
        )


class MembershipType(Auditable, models.Model):
    name = models.CharField(max_length=200, verbose_name=_("name"))
    amount = models.DecimalField(
        max_digits=8,
        decimal_places=2,
        verbose_name=_("amount"),
        help_text=_("Please enter the yearly fee for this membership type."),
    )


class Membership(Auditable, models.Model, LogTargetMixin):
    LOG_TARGET_BASE = "byro.members.membership"

    member = models.ForeignKey(
        to="members.Member", on_delete=models.CASCADE, related_name="memberships"
    )
    start = models.DateField(verbose_name=_("start"))
    end = models.DateField(verbose_name=_("end"), null=True, blank=True)
    amount = models.DecimalField(
        max_digits=8,
        decimal_places=2,
        verbose_name=_("membership fee"),
        help_text=_("The amount to be paid in the chosen interval"),
    )
    interval = models.IntegerField(
        choices=FeeIntervals.choices,
        verbose_name=_("payment interval"),
        help_text=_("How often does the member pay their fees?"),
    )

    form_title = _("Membership")

    def get_absolute_url(self):
        return reverse("office:members.data", kwargs={"pk": self.member.pk})

    def get_dues(self, _now=None, _from=None):
        _now = _now or now()
        dues = set()
        end = self.end
        start = self.start
        if _from is not None and start < _from:
            start = _from
        if not end:
            try:
                end = _now.replace(day=start.day).date()
            except (
                ValueError
            ):  # membership.start.day is not a valid date in our month, we'll use the last date instead
                end = (_now + relativedelta(day=1, months=1, days=-1)).date()
        date = start
        while date <= end:
            dues.add((date, self.amount))
            date += relativedelta(months=self.interval)
        return (start, end), dues


SPECIAL_NAMES = {Member: "member", Membership: "membership"}

SPECIAL_ORDER = [
    "member__number",
    "member__last_name",
    "member__first_name",
    "member__address",
    "member__email",
    "membership__start",
    "membership__interval",
    "membership__amount",
]


class MemberBalance(models.Model):
    """Member balance entries are similar in nature to invoices, describing the
    amount owed over a certain period in time.

    As they MUST NOT overlap per member, they should be created via the
    ``Member.create_balance(start, end)`` method interface.
    """

    member = models.ForeignKey(
        to="members.Member", related_name="balances", on_delete=models.PROTECT
    )
    reference = models.CharField(
        null=True,
        blank=True,
        verbose_name=_("Reference"),
        help_text=_("For example an invoice number or a payment reference"),
        unique=True,
        max_length=50,
    )
    amount = models.DecimalField(
        max_digits=8, decimal_places=2, verbose_name=_("Amount")
    )
    start = models.DateTimeField(verbose_name=_("Start"))
    end = models.DateTimeField(verbose_name=_("End"))
    state = models.CharField(
        choices=[
            ("paid", _("paid")),
            ("partial", _("partially paid")),
            ("unpaid", _("unpaid")),
        ],
        default="unpaid",
        max_length=7,
    )



import os
from itertools import repeat


def log_initial(*, debug, config_files, db_name, LOG_DIR, plugins):
    from byro import __version__
    from byro.common.console import end_box, print_line, start_box

    if hasattr(os, "geteuid") and os.geteuid() == 0:
        print_line("You are running byro as root, why?", bold=True)

    mode = "development" if debug else "production"
    lines = [
        (
            "This is byro {__version__} calling, running in {mode} mode.".format(
                __version__=__version__, mode=mode
            ),
            True,
        ),
        ("", False),
        ("Settings:", True),
        ("Read from: " + ", ".join(config_files), False),
        (f"Logging:   {LOG_DIR}", False),
    ]
    if plugins:
        lines += [("Plugins:   " + ",".join(plugins), False)]
    else:
        lines += [("", False)]
    image = """
┏━o━━━━o━━━┓
┣━━━o━━━o━━┫
┣━━━━━━━━━━┫
┃   byro   ┃
┗━━━━━━━━━━┛
    """.strip().split(
        "\n"
    )
    img_width = len(image[0])
    image[-1] += " " * (img_width - len(image[-1]))
    image += [" " * img_width for _ in repeat(None, (len(lines) - len(image)))]

    lines = [(image[n] + " " + line[0], line[1]) for n, line in enumerate(lines)]

    size = max(len(line[0]) for line in lines) + 4
    start_box(size)
    for line in lines:
        print_line(line[0], box=True, bold=line[1], size=size)
    end_box(size)


def reduce_dict(data):
    return {
        section_name: {
            key: value for key, value in section_content.items() if value is not None
        }
        for section_name, section_content in data.items()
    }
